!
!  Module      : omatt_mod
!
! Authors     
!             : Vladimir Yershov (MSSL)  
!             : Chris Brindle (MSSL)  
!             : Tim Harries
!             : Andrew King (MSSL)
! Description : Convert from OM detector to sky coordinates and perform sky aspect-correction, if possible.
! 
! Comments    : 
!
! $Id: module.f90,v 1.4 1997/12/18 16:01:10 gvacanti Exp $
! 
! History : Created  : March 98 - TJH
!           Modified : 21 May 98 - Added use sla.
!           Modified : 27 May 99 - Updated parameter names.
!           Modified : 1st June 00 - Added calculation of galactic coordinates for RA and DEC
!           Modified : 1st June 00 - Added addition of two extra keywords and columns to hold
!                                    the galactic coordinates.
!           Modified : 2nd June 00 - Changed the galactic coordinates to 8 byte data
!           Modified : 2nd June 00 - Installed code to release memory that had been allocated
!                                    this was left hanging allocated.
!           Modified : 4th june 00 - Changed RA and DEC to 8 byte data
!
!
!           Modified : 21/06/2000 Changed poserr to be consistent with arcsecs definition from OMDETECT.
!
!           Modified : 1/8/2000  (CB) - Added -90 degree image rotation
!                      to correct for error in coordinate system.
!           Modified : 14/8/2000 (CB) - Maths functions sin and cos replaced by dsin and 
!                                       dcos (double precision versions),
!                                       & some other changes to double precision made.
!           Modified : 15/8/2000 (CB) - Output image corrected for image distortion.
!           Modified : 21/8/2000 (CB) - Function "MissionTime" added to compute the difference
!                                       between the average of the start and end observations
!                                       currently fixed at 0.4765
!          Modified : 24/8/2000 (CB) -  Some write statements deleted. The +ve signs for 
!                                       the distortion corrections changed to -ve.
!          Modified : 07/09/2000     -  Attitude matrinningix read in. Boresight calculation changed -
!                                       uses [A*(B)^-1]*(1,0,0).
!          Modified : 28/09/2000(AMK)-  Changed parameters in call message!
!          modified : 02/10/2000(AMK)-  Reinstated the call to CAL_omGetPLateScalePnte Rotate
!
!          modified : 02/10/2000(SRR)-  Inserted inv_boremat for the inverted boresight matrix
!
!          modified : 05/10/2000(AMK)-  Replaced the call getFIlterID in mssllib by the OAL call filterIdToInt
!          modified : 10/10/2000(CB) -  Paramter "usecat" set to false if it is true- not yet implemented
!          modified : 20/12/2000(CB) -  Removed world-coordinate attributes from input image header file.
!          modified : 18/01/2000(CB) -  Exposure array added to the PPS output file. Code for astrometry
!                                       correction moved to a subroutine. 
!          modified : 24/05/2001(CB) -  Attributes "RA_SCX, DEC_SCX & PA added to output file- values
!                                       refer to the spacecraft pointing position. RA_PNT & RA_DEC
!                                       also added.  
!          modified : 14/06/2001(CB) -  Removed code that changes position errors (SPR 2429)
!          modified : 02/08/2001(CB) -  Removed definition of struct yySpaceCraftAttitudeType-
!                                       defined in caloaldefs.  Replaced the matrix attmatrix in the 
!                                       call to OAL_toAttitudeMatrix by attmatrix1 (correct subscripts)
!                                       Changed program so that all 1-d arrays start at 1. Tidied
!                                       program up- removed astrometry correction code into a
!                                       separate module- added module "utility" which contains
!                                       some subroutines for doing routine jobs. 
! 
!          modified : 10.09.2001(VNY) - The centre of the CCD is fixed at (1024.5, 1024.5). Output RA and 
!                                       DEC coordinates has been recalculated according to the origin of the
!                                       CCD coordinate system lying at the center of the first pixel.
!                                       Correspondingly, the output image shift and rotation have
!                                       been corrected                            
!          modified : 16.11.2001 (CB) - Code to store an exposure-image map removed. Rotated image pixels
!                                       outside the image boundary set to NULL
!          modified : 16/12/2001(CB) -  Parameter "usecat" reactivated and astrometry correction now done
!                                       if sufficient stars   
!          modified : 17/09/2002 (VNY) - adjusted to handle with the source list without the MAG-column 
!                                       (omdetect now does not generate it). If there is no MAG 
!                                       column in the SRCLIST-file, the magnitudes are set to zero 
!
!         modified : 25/09/2002 (CB) - Position angles of each source adjusted for image rotation 
!         modified : 05/11/2002 (CB) - pgplot code removed
!         modified : 02/12/2002 (CB) - added boolean parameter to control production of the 
!                                      rotated image - (false no image- defaults to true) 
!         07.06.2006 (VNY) The scale factor for the sky-grism-image is corrected  
!         14.06.2006 (VNY) Eliminating the duplicated keywords CTYPE!, CTYPE2 in the header 
!         27.07.2006 (VNY) Computing the unrotated source positions (modifying the subroutine 
!         ComputeGrismCelestialCoordinates)
!         19.09.2006 (VNY) checking the problem with the keywords CRVAL1 and CRVAL2
!         20.09.2006 (VNY) adjusting the zero-point of the RA and DEC for the grism sources
!         21.09.2006 (VNY) Using the modified DETSEC value for calculating the source
!                          coordinates
!         05.10.2006 (VNY) Adjusting the position of the sky-image for grisms
!         07.11.2006 (VNY) Introducing offsets for grism filters (different for 
!                          different resolutions (subroutine ComputeGrismCoordinates2)
!         09.11.2006 (VNY) Last corrections of the offsets for grisms and cleaning
!                          from the test messages> Changing the version number
!                          from 2.41 to 2.42 (SSC-SPR-3633)
!         13.02.2007 (VNY) introducing the correction for the Full frame-low res, prim instr. R1
!         28.02.2007 (VNY) Making the default prime instrument R1 instead of OM (this is 
!                          for the case when the prime onstr. information is missing in the 
!                          SAS summary file (early observations)
!         01.03.2007 (VNY) Finished calibrations of GetGrismPosition
!         02.03.2007 (VNY) Removing the test messages
!         15.02.2008 (VNY) Adding the astrometric corrections to the 
!                          CRVAL1 and CRVAL2 values of the rotated (sky) image 
!                          in the case of the parameter "usecat=yes" (when the 
!                          astrometric correction is calculated)
!         07.04.2008 (VNY) Added two new attributes to the rotated (sky) image 
!                FITs file: RA_OFF and DEC_OFF (in arcseconds) to indicate 
!                the amount of astrometric correction applied to the image
!         21.04.2008 (VNY) The RA_OFF is divided by cos(dec) in order to get the
!                          correct CRVA1 for the mosaiced image (further improvement
!                          related to SSC-SPR-3710
!         16.06.2008 (VNY) A bug was fixed, which was causing the setting of the 
!                          POSCOROK keyword in the header of the rotated image
!                          to TRUE when the number of matching sources was too small
!                          and the actual value of this parameter must be FALSE
!                          (SSC-SPR-6478).
!         03.02.2012 (ATI) Adapted for using time variable boresight.
!                          Subroutine  OmPointingDirection modified.                            
!                          Subroutine  OmGrismPointingDirection modified (SSC-SPR=6690). 
!         23.10.2012 (VNY) Added the functionality to generate its own subset of the USNO
!                          catalogue if it is not among the input files
!*****************************************************************************************  
module omatt_mod

  use param
  use oal
  use cal
  use dal
  use sasdal
  use caloaldefs

  use errorHandling
  use sla
  use readModes_mod
  use omallocation_mod 


  use omconstants_mod
!  use omastromcorrect_mod    ! For doing the astrometry correction
 
  use integration_mod
  use omdistortion_mod
  use omrotation_mod
  use omcelestial_mod
  use omsortra_mod
implicit none


public
   public omatt

contains

!*******************************************************************
! subroutine omatt
! NB Paramters xCentre & yCentre are stored in module omconstants_mod
! in mssllib and are 1024.5 * 1024.5
! NB the input pixel coordinates are defined such that the centre
! of the bottom-left pixel is 1, 1
!******************************************************************
subroutine omatt


    type(DataSetT) :: inDataSet                                       ! Input dataset handle.
    type(DataSetT) :: sourceListDataSetH                              ! Source List dataset handle.
  
    type(TableT) :: sourceListTableH                                  ! Source List table handle.

    character(255) :: sourceListSet                               ! OM OSW Source List filename
    character(10)           :: datamode   ! datamode attribute
    character(6)            :: omFilter   ! the OM filter
    integer                 :: xBinning
    integer                 :: yBinning
    integer             :: nrows              ! the number of rows in the table
    integer             :: xbin, ybin         ! binning factors
    integer             :: xsize, ysize       ! size of the image
    integer                     :: OM_MODE    ! the om mode
    real(double)  :: rotationCorrection = 90.0_double! Correction for coordinate error
    real(double)  :: plateScale       ! the platescale
    real(double) :: ra_point = 0.0_double, dec_point= 0.0_double        
    ! position of the OM tangent point
    
    real(double) :: cosDec
               
    real(single), dimension(:), pointer  ::  xpos => null()
    real(single), dimension(:), pointer  ::  ypos => null()

    real(double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     
    real(double), dimension(:), pointer :: rightAscension => null()
    real(double), dimension(:), pointer :: declination => null()
    real(single), dimension(:), pointer :: posErr => null()
    real(double), pointer :: galacticLong(:) => null()  ! Array for galactic longitudes  
    real(double), pointer :: galacticLat(:)  => null()  ! Array for galactic latitudes
  
    real(single), dimension(:), pointer :: magnitude => null()
    real(double) :: rollangle                  ! the roll angle
    real(double) :: grismRollangle             ! the grism filter roll angle                        
  
    real(double)  :: lambda           ! wavelength for CAL
    real(single)  :: nullValue
    logical(Bool) :: binBPE                    ! BPE binning on?
    logical(bool) :: useCat                        ! Use the USNO-SA 1 catalog.
    
    integer                  :: istat              ! status for memory allocation
    integer(int32)           :: filterIdInt
    
    integer(int32) :: position
 
    character(255) :: set                                         ! OM image intermediate filename
    integer(int32) :: i, x01, y01
    character(len=7)            :: tempFilter   ! temporary the OM filter
    logical(bool) :: grismObservation
    real(double) :: omgXmin, omgYmin
    real(double) :: omgXref, omgYref
    
    character(len=255) :: detsecString
    character(len=1) :: ch1, ch2, ch3, ch4, ch5
    integer(kind=int32) :: windowX0, windowX1, windowY0, windowY1
    real(double) :: grismOffsetX, grismOffsetY
    character(len=10) :: primInstrValue
    type(CorrectionData) :: corrData
    logical(bool) :: posCorr = .false. 
    logical :: corrWasOkay  =  .false.

    nullValue = real(-99999)
    omgXref=1024.5 ! setting these variable just in case the header keyword is missing
    omgYref=1024.5 
    grismObservation=.false.
    ! initial values for grismOffsetX and grismOffsetY (valid for full frale high resolution)
    grismOffsetY=-9.72_double
    grismOffsetX=0.0_double
    posCorr=.false.

    !*************************************
    ! Get the command-line parameters
    !*************************************
    set = stringParameter("set") ! This is the input image (rotated image in the case of Grism)
    inDataSet = DataSet(set, read) 

    !***********************************************************
    ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
    !***********************************************************
    call readModes(inDataSet, x01, y01, xsize, ysize)
    x0 = x01
    y0 = y01   

    !********************************************************************
    ! Read in the DETSEC values (they correspond to the unrotated image
    !********************************************************************
    if (hasAttribute(inDataSet,"DETSEC")) then
       detsecString = stringAttribute(inDataSet, "DETSEC")
       read (detsecString, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
            ch1,windowX0, ch2, windowX1, ch3,  windowY0, ch4, windowY1, ch5
    endif

    sourceListSet = stringParameter("sourcelistset")
    useCat = booleanParameter("usecat")
    
    !**************************
    ! Access the source list
    !**************************
    sourceListDataSetH = dataSet(sourceListSet, modify)      ! get osw list handle
   
    xbin = int32Attribute(sourceListDataSetH, "BINAX1")      ! binning factors
    ybin = int32Attribute(sourceListDataSetH, "BINAX2")
 
    if (xbin .ne. ybin) then
        call fatal("diffBin", "This version cannot deal with different binning factors in x and y")
    end if

    binBPE = booleanAttribute(sourceListDataSetH, "BINBPE")
 
    ! Calculate the image binning.
    xBinning = 2 ** xbin
    yBinning = 2 ** ybin


    if (binBPE) then
        xBinning = xBinning * 2
        yBinning = yBinning * 2
    end if


    dXbinning = dble(xBinning)  ! Need the double values of xBinning and 
    dYbinning = dble(yBinning)  ! yBinning fairly frequently
   
    call InitialiseDistortion(x01, y01, xBinning, yBinning)

!------------  13.02.2006 ---------------
    tempFilter="GRISM10"
    omFilter = stringAttribute(sourceListDataSetH, "FILTER")
    if (trim(omFilter) .eq. "Grism1" .or. trim(omFilter) .eq. "GRISM1") then 
       omFilter="GRISM1"
       tempFilter="GRISM10"
       grismObservation=.true.
    endif
    if (trim(omFilter) .eq. "Grism2" .or. trim(omFilter) .eq. "GRISM2") then
       omFilter="GRISM2"
       tempFilter="GRISM20"
       grismObservation=.true.
    endif

    ! Visual grism rollangle (Grism2)
    if (omFilter .eq. "GRISM2  ") grismRollangle=-5.6
    if (omFilter .eq. "GRISM1  ") grismRollangle=6.3

    omgXmin=0.0_double
    omgYmin=0.0_double
    if (grismObservation) then
       if (hasAttribute(inDataSet, "OMGXMIN")) then
          omgXmin = real64Attribute(inDataSet, "OMGXMIN")
       else
          call warning("OM grism image","Header keyword IMGXMIN not found")
       endif
       if (hasAttribute(inDataSet, "OMGXMIN")) then      
          omgYmin = real64Attribute(inDataSet, "OMGYMIN")
       else
          call warning("OM grism image","Header keyword IMGYMIN not found")
       endif
       if (hasAttribute(inDataSet, "CRPIX1")) then      
          omgXref = real64Attribute(inDataSet, "CRPIX1")
       else
          call warning("OM grism image","Header keyword CRPIX1 not found")
       endif
       if (hasAttribute(inDataSet, "CRPIX2")) then      
          omgYref = real64Attribute(inDataSet, "CRPIX2")
       else
          call warning("OM grism image","Header keyword CRPIX2 not found")
       endif
    endif
!--------------------------------------------
    dataMode = stringAttribute(sourceListDataSetH, "DATAMODE")  ! needed for cal
    if (trim(dataMode) .eq. "IMAGING") then 
       OM_MODE = IMAGE
    else
       OM_MODE = FAST
    end if
   
    sourceListTableH = table(sourceListDataSetH, "SRCLIST")     ! get the source list

    nrows = numberOfRows(sourceListTableH)

    !*********************************************************
    ! Get the x and y pixel coordinates of each source
    ! NB these are stored as real32 values.
    !********************************************************

    xPos => allocateSingle(1, nRows, "xpos")
 
    call ReadReal32Column(sourceListTableH, "XPOS", nRows, 0, xpos)

    yPos => allocateSingle(1, nRows, "ypos")
  
    call ReadReal32Column(sourceListTableH, "YPOS", nRows, 0, ypos)
  

    magnitude => AllocateSingle(1, nRows, "magnitude")
    poserr => allocateSingle(1, nRows, "poserr")
   

    !----------------- VNY 17.09.2002 Checking if the MAG column does exist 
    if (hasColumn(sourceListTableH, "MAG")) then
        call ReadReal32Column(sourceListTableH, "MAG", nRows, 0, posErr)
        magnitude = real(posErr)      
    else 
       magnitude = 0.0
    endif
    !----------------------
 
  
    call ReadReal32Column(sourceListTableH, "POSERR", nRows, 0, posErr)
 
 
    !*******************************************************
    ! get the astrometry info keywords from the source list
    !*******************************************************
    ra_scx = real64Attribute(sourceListDataSetH, "RA_SCX")   ! Spacecraft RA in radians
    dec_scx = real64Attribute(sourceListDataSetH, "DEC_SCX") ! Spacecraft dec in radians

    pa_scx = real64Attribute(sourceListDataSetH, "PA")       ! Spacecraft PA in degrees
   
    ra_scx = ra_scx * degreesToRadians
    dec_scx = dec_scx * degreesToRadians
 
    !********************************************
    ! Transform the source position angles
    ! to the equatorial coordinate system
    !*******************************************
    if (.not. grismObservation) then
       call ModifySourcePositionAngles(sourceListTableH, pa_scx + rotationCorrection, nullValue)
    endif
    pa_scx = pa_scx * degreesToRadians
    rollAngle = pa_scx

    !**********************************************
    ! rotationCorrection corrects for OM axes definition (90 degrees).
    !**********************************************
    rotationCorrection = rotationCorrection * degreesToRadians
    rollAngle = rollAngle + rotationCorrection
    if (grismObservation) then!

       ! VNY 20.09.2006 the true roll angle is that, when
       ! the spectra were not aligned with the y-axis
       ! 02.11.2006 multiplying the grismRollAngle by 2
       !  since it is seemed overrotating
       !rollAngle=rollAngle-grismRollAngle*degreesToRadians
       rollAngle=rollAngle-2.0*grismRollAngle*degreesToRadians
       
    endif

    !************************************
    ! Get the filter Id integer value
    !************************************
    if (trim(omFilter) .eq. "Grism1" .or. trim(omFilter) &
         .eq. "GRISM1" .or. trim(omFilter) .eq. "Grism2" &
         .or. trim(omFilter) .eq. "GRISM2") then
       !tempFilter="GRISM10"
       filterIdInt = filterIdToInt(tempFilter)
    else
       filterIdInt = filterIdToInt(omFilter)
    endif

    !*********************
    ! Set the CAL state
    !*********************
    call CAL_setState(OM, ccdModeId=OM_MODE, filterId = filterIdInt)


    !******************************************
    ! Boresight stuff
    !******************************************
    if (grismObservation) then
       call OmGrismPointingDirection(inDataSet, ra_point, dec_point,grismRollAngle)
    else
       call OmPointingDirection(inDataSet, ra_point, dec_point)
    endif

    call release(inDataSet)

    call CAL_omGetPlateScale(lambda, plateScale)  ! Get the plate-scale (arcsecs per pixel) 

    plateScale = plateScale * degreesToRadians / 3600.0_double    
    ! platescale converted to radians
  
    sinRollAngle = dsin(rollAngle)  ! Need these quantities quite a lot
    cosRollAngle = dcos(rollAngle)

    rightAscension => AllocateDouble(1, nRows, "rightAscension")
    declination => AllocateDouble(1, nRows, "declination")
    galacticLong => AllocateDouble(1, nrows, "galacticLong")
    galacticLat => AllocateDouble(1, nrows, "galacticLat")
    if (grismObservation) then
       primInstrValue='' 
       call GetPrimaryInstrument(set,primInstrValue)    
       call GetGrismPosition(primInstrValue, omFilter, binBpe, xBinning, &
            yBinning, grismOffsetX, grismOffsetY)
    endif

    !*********************************************************************
    ! For each source compute its undistorted position 
    ! relative to the centre of the full-frame image, and then
    ! its RA, dec, galactic longitude and latitude
    !*********************************************************************
    do i = 1, nrows
!************************************
       if (grismObservation) then   
          call ComputeGrismCoordinates2(i, xPos(i), yPos(i), x0, y0,   & 
               real(xBinning), real(yBinning), &
               plateScale, rollAngle, grismRollAngle, &
               ra_point, dec_point, &
               rightAscension(i), declination(i) , &
               galacticLong(i), galacticLat(i),&
               omgXmin,omgYmin,windowX0, windowY0, &
               grismOffsetX, grismOffsetY)
       else
!************************************
          call ComputeCelestialCoordinates(xPos(i), yPos(i), x0, y0,   & 
               real(xBinning), real(yBinning), &
               plateScale, rollAngle, &
               ra_point, dec_point, &
               rightAscension(i), declination(i) , &
               galacticLong(i), galacticLat(i))
       endif
    end do

    !*********************************************************
    ! Add RA, dec, galactic longitude and galactic latitude 
    ! columns to the fits source-list file
    !*********************************************************
    position = columnNumber(sourceListTableH, "POSERR") + 1

    call AddReal64Column(sourceListTableH, rightAscension * radiansToDegrees, &
             nrows, "RA", "degrees", "Not corrected for boresight offset", nullvalue, position)
    call AddReal64Column(sourceListTableH, declination * radiansToDegrees, &
             nrows, "DEC", "degrees", "Not corrected for boresight offset", nullValue, position + 1)
    call AddReal64Column(sourceListTableH, galacticLong  * radiansToDegrees, &
                   nrows, "LII", "degrees", "Not corrected for boresight offset", nullValue, position + 2)
    call AddReal64Column(sourceListTableH, galacticLat  * radiansToDegrees, &
                    nrows, "BII", "degrees", "Not corrected for boresight offset", nullValue, position + 3)


    
    !****************************************************************
    !
    ! We only do the next bit if we have access to the usno catalogue
    !
    !****************************************************************    
    if (useCat) then    
        !************************************************************
        ! NB all angles should be in radians, apart from poserr
        ! which is in arc secs
        !************************************************************
        call CorrectTheAstrometry(nrows,  rightAscension, declination, &
            posErr, magnitude, ra_point, dec_point, sourceListDataSeth, &
            sourceListTableH, corrData, corrWasOkay) 
        posCorr = corrWasOkay

    end if



    deallocate(rightAscension, declination, posErr,                 &
                magnitude, galacticLong, galacticLat, stat = istat)
    call CheckDeallocation(istat, "rightAscension")
 
    call SetFitsTableAttributes(sourceListDataSetH, ra_point, dec_point)


    if( .not. grismObservation) then
        call sortTable( sourceListTableH )
    end if

    call release(sourceListTableH)
    call release(sourceListDataSetH)

    cosDec= dcos(dec_point)
    
    if (booleanParameter("rotateimage")) then  ! Create the rotated image
       if(grismObservation) then
          call CreateGrismSkyImage(trim(stringParameter("set")), &
               trim(stringParameter("ppsoswset")), &
               rollAngle, grismRollAngle, ra_point, dec_point, &
               xBinning, yBinning, ra_scx, dec_scx, plateScale, &
               omgXref, omgYref)
        else          
           if (useCat) then
              ! VNY 21.04.2008 xShift must be divided by cos(dec)
              corrData%xShift=corrData%xShift/cosDec
              ra_point=ra_point+corrData%xShift*degreesToRadians/3600.0
!              ra_point=ra_point+corrData%xShift*degreesToRadians/3600.0/cosDec
              dec_point=dec_point+corrData%yShift*degreesToRadians/3600.0
              write(errstr, '(a, f6.2,a,f6.2,a)') "Adding astrometry correction to the image (", &
                   corrData%xShift,"''",corrData%yShift, "'')" 
              call message(AppMsg, Verbose)
              !posCorr=.true.
              posCorr = corrWasOkay
           endif
           ! Rotate the image by rollAngle (radians)

!write(*, *) "Line 561 posCorr=", posCorr

           call CreateRotatedImage(trim(stringParameter("set")), &
                trim(stringParameter("ppsoswset")), &
                rollAngle, ra_point, dec_point, &
                xBinning, yBinning, ra_scx, dec_scx, &
                plateScale, posCorr, corrData%xShift, corrData%yShift)

           !*********************************************************************
           ! Added by CB on 18th July 2008
           !
           !********************************************************************
           call addQualityImage( trim(stringParameter("set")),        &
                                 trim(stringParameter("ppsoswset")),  &
                                 rollAngle, ra_point, dec_point,      &
                                 xBinning, yBinning, ra_scx, dec_scx, &
                                 plateScale, grismObservation)

        endif
    end if

    !**************************************
    ! Release memory allocation
    !**************************************
    
    deallocate(rawx, rawy, xpos, ypos, stat = istat)
    call CheckDeallocation(istat, "xpos")

 
end subroutine omatt

!*******************************************
! Function MissionTime - Returns the number
! of seconds from the start of the mission
! to the observation time. Added by CB because
! the call to CAL_GetBoresiteMatrix needs
! the time in seconds from the launch.
!******************************************
real(double) function  MissionTime(primaryTable)
    type(DataSetT), intent(in)   :: primaryTable

    character(19)  :: dateObs         

    real(double)  :: JulianDate0, JulianDate1 ! , JulianDate2
    character(19)  :: missionReferenceTime = "1998-01-01T00:00:00"

    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the start of the observations.
    dateObs = stringAttribute(primaryTable,  "DATE-OBS")    
    JulianDate1 = ModifiedJulianDate(dateObs)


   MissionTime = (JulianDate1 - JulianDate0) * 86400.0d0


end function MissionTime


!********************************************
! Subroutine ModifySourcePositionAngles
!*******************************************
subroutine ModifySourcePositionAngles(tableSetH, angle, nullValue)
    type(TableT), intent(in) :: tableSetH
    real(double), intent(in) :: angle
    real(single), intent(in) :: nullValue
    
    real(single), pointer :: positionAngle(:) => null()
    integer(int32) :: i, noRows, position
  
    noRows = numberOfRows(tableSetH)

    position = columnNumber(tableSetH, "PA")
 
    positionAngle => allocateSingle(1, noRows, "positionAngle")
  
    call ReadReal32Column(tableSetH, "PA", noRows, 0, positionAngle)
 

 
    call deleteColumn(tableSetH, "PA")


    do i = 1, noRows
        if (.not. isNullValue(positionAngle(i))) then 
            positionAngle(i) = positionAngle(i) + angle
            if (positionAngle(i) .gt. 360.0) positionAngle(i) = positionAngle(i) - 360.0
            if (positionAngle(i) .gt. 180.0) positionAngle(i) = positionAngle(i) - 180.0
        end if
    end do
    call AddReal32Column(tableSetH, positionAngle, noRows, "PA", "degrees", &
             "PA of major-axis Referred to equatorial system (measured anti-clockwise)", nullvalue, position)
 

    deallocate(positionAngle)
end subroutine ModifySourcePositionAngles

!***************************************************************
!
!**************************************************************
subroutine OmPointing(dataSet)
    type(dataSetT), intent(in) :: dataSet

    real(double) :: ra_point, dec_point        ! position of the OM tangent point
    real(double) :: rollangle                  ! the roll angle
    real(double) :: time                        
    real(double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
    real(double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

    real(double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
    real(double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
    real(double) :: pointvec(3)        ! s/c pointing
    real(double) :: pointvec2(3)       
    real(double), pointer :: boremat(:,:)       ! boresight offset matrix
    real(double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
    real(double) :: omvec(3)                   ! om pointing vector
  

    real(double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     
    real(double) :: r1                         ! modulus
    type (SpacecraftAttitudeType) attitude
     
    
    integer i, j    

    !***********************************************************
    ! Get the number of seconds since the mission reference time
    !***********************************************************
    time =  MissionTime(dataSet)


    !*******************************************************
    ! get the astrometry info keywords from the source list
    !*******************************************************
    ra_scx = real64Attribute(dataSet, "RA_SCX") * degreesToRadians   ! Spacecraft RA in radians
    dec_scx = real64Attribute(dataSet, "DEC_SCX") * degreesToRadians ! Spacecraft dec in radians
    pa_scx = real64Attribute(dataSet, "PA")        ! Spacecraft PA in degrees
 
 
    pa_scx = pa_scx * degreesToRadians
    rollAngle = pa_scx

    
    call CAL_GetBoresightMatrix(time, boremat) ! boresight offset matrix
    
    !*******************************************************************
    ! need inverted boresight matrix which is obtained via the transpose. 
    !********************************************************************
    inv_boremat = TRANSPOSE(boremat)

    attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle) 

    !*******************************************************
    ! This routine requires a matrix of size 0:2, 0:2, but
    ! after that we want it to be of size 1:3, 1:3
    !******************************************************
    call OAL_toAttitudeMatrix(attitude, attmatrix1)
    
    do i = 1,  3 
        do j = 1, 3 
            attmatrix(i, j) = attmatrix1(i - 1, j - 1)
        end do
    end do

    !******************************************************************************
    ! converts RA and DEC from spherical cordinates to cartesian coordinates vector
    !*****************************************************************************
    call sla_dcs2c(ra_scx, dec_scx, pointvec)

    !********************************************************
    ! normalize the vector pointvec and return in pointvec2 -
    ! r1 holds the mdulus of the vector. pointvec2 will store
    ! the unit-vector in the direction of pointvec
    !********************************************************
    call sla_dvn(pointvec, pointvec2, r1)

    !
    !
    ! use SIMON'S BIT
    !
    !write(*,*)'using SIMONS BIT (essentially redundant now but keep for now
    !------------------
    pointvec(1) = -pointvec2(1)
    pointvec(2) = -pointvec2(2)
    pointvec(3) =  pointvec2(3)
    !------------------------------------


    ! convert euler axis vector to a rotation matrix
     

    ! initialise output vector
    omvec = 0.0_double


!**********************************************************************************************
!  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
!  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
!  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
!
!  This creates omvec, the om boresight vector measured in the ERF
!  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
!  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
!  is the OM boresight as measured in the OM reference frame
!*********************************************************************************************

    att_invbor = matmul(attmatrix, inv_boremat)
    omvec = matmul(att_invbor, ompoint)

    !************************************************************************
    ! converts the omvec cartesian vector to ra and dec spherical coordinates
    ! and check RA lies within sensible range
    ! ra_point and ra_dec refer to the pointing direction of the instrument
    !************************************************************************

    call sla_dcc2s(omvec, ra_point, dec_point)

    if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi
   
    write(errstr, '(a, f8.4)') "RA of OM boresight: ", ra_point * radiansToDegrees
    call message(AppMsg, Verbose)
    write(errstr, '(a, f8.4)') "DEC of OM boresight: ", dec_point * radiansToDegrees
    call message(AppMsg, Verbose)


end subroutine OmPointing




subroutine OmPointing1(dataSet)
    type(dataSetT), intent(in) :: dataSet

    real(double) :: ra_point, dec_point        ! position of the OM tangent point
    real(double) :: rollangle                  ! the roll angle
   
    real(double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
    real(double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

    real(double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
    real(double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
    real(double) :: pointvec(3)        ! s/c pointing
    real(double) :: pointvec2(3)       
    real(double), pointer :: boremat(:,:)       ! boresight offset matrix
    real(double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
    real(double) :: omvec(3)                   ! om pointing vector

    real(double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     
    real(double) :: r1                         ! modulus
    type (SpacecraftAttitudeType) attitude
     
    integer i, j    
    real(double) :: time


    !*******************************************************
    ! get the astrometry info keywords from the source list
    !*******************************************************
    ra_scx = real64Attribute(dataSet, "RA_SCX") * degreesToRadians   ! Spacecraft RA in radians
    dec_scx = real64Attribute(dataSet, "DEC_SCX") * degreesToRadians ! Spacecraft dec in radians
    pa_scx = real64Attribute(dataSet, "PA")        ! Spacecraft PA in degrees
 
 
    pa_scx = pa_scx * degreesToRadians
    rollAngle = pa_scx



    !***********************************************************
    ! Get the number of seconds since the mission reference time
    !***********************************************************
    time =  MissionTime(dataSet)

   

    call CAL_GetBoresightMatrix(time, boremat) ! boresight offset matrix
   
    
    !*******************************************************************
    ! need inverted boresight matrix which is obtained via the transpose. 
    !********************************************************************
    inv_boremat = TRANSPOSE(boremat)

    attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle)

    !*******************************************************
    ! This routine requires a matrix of size 0:2, 0:2, but
    ! after that we want it to be of size 1:3, 1:3
    !******************************************************
    call OAL_toAttitudeMatrix(attitude, attmatrix1)
    
    forall( i = 1 : 3) 
        forall(j = 1 : 3) attmatrix(i, j) = attmatrix1(i - 1, j - 1)
    end forall

   
 
    !******************************************************************************
    ! converts RA and DEC from spherical cordinates to cartesian coordinates vector
    !*****************************************************************************
    call sla_dcs2c(ra_scx, dec_scx, pointvec)

    !********************************************************
    ! normalize the vector pointvec and return in pointvec2 -
    ! r1 holds the mdulus of the vector. pointvec2 will store
    ! the unit-vector in the direction of pointvec
    !********************************************************
    call sla_dvn(pointvec, pointvec2, r1)

    !
    !
    ! use SIMON'S BIT
    !
    !write(*,*)'using SIMONS BIT (essentially redundant now but keep for now
    !------------------------------------------------
    pointvec(1) = -pointvec2(1)
    pointvec(2) = -pointvec2(2)
    pointvec(3) =  pointvec2(3)
    !-------------------------------------------------

 

    ! convert euler axis vector to a rotation matrix
     

    ! initialise output vector
    omvec = 0.0_double


!**********************************************************************************************
!  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
!  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
!  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
!
!  This creates omvec, the om boresight vector measured in the ERF
!  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
!  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
!  is the OM boresight as measured in the OM reference frame
!*********************************************************************************************

    att_invbor = matmul(attmatrix, inv_boremat)
    omvec = matmul(att_invbor, ompoint)

    !************************************************************************
    ! converts the omvec cartesian vector to ra and dec spherical coordinates
    ! and check RA lies within sensible range
    ! ra_point and ra_dec refer to the pointing direction of the instrument
    !************************************************************************

    call sla_dcc2s(omvec, ra_point, dec_point)

    if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi
   
  

    write(errstr, '(a, f8.4)') "RA of OM boresight: ", ra_point * radiansToDegrees
    call message(AppMsg, Verbose)
    write(errstr, '(a, f8.4)') "DEC of OM boresight: ", dec_point * radiansToDegrees
    call message(AppMsg, Verbose)


end subroutine OmPointing1


!****************************************************************
!
!
!***************************************************************
subroutine OmPointingDirection(dataSet, ra_point, dec_point)
    type(dataSetT), intent(in) :: dataSet

    real(double), intent(out) :: ra_point, dec_point  ! position of the OM tangent point
    real(double) :: rollangle                  ! the roll angle
   
    real(double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
    real(double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

    real(double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
    real(double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
    real(double) :: pointvec(3)        ! s/c pointing
    real(double) :: pointvec2(3)       
    real(double), pointer :: boremat(:,:)       ! boresight offset matrix
    real(double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
    real(double) :: omvec(3)                   ! om pointing vector
  

    real(double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     
    real(double) :: r1                         ! modulus
    type (SpacecraftAttitudeType) attitude
     
    type(BoresightT) :: boresight    
    integer i, j    
    real(double) :: time

    ra_point = 0.0_double
    dec_point = 0.0_double


    !*******************************************************
    ! get the astrometry info keywords from the source list
    !*******************************************************
    ra_scx = real64Attribute(dataSet, "RA_SCX") * degreesToRadians   ! Spacecraft RA in radians
    dec_scx = real64Attribute(dataSet, "DEC_SCX") * degreesToRadians ! Spacecraft dec in radians
    pa_scx = real64Attribute(dataSet, "PA")        ! Spacecraft PA in degrees
 
 
    pa_scx = pa_scx * degreesToRadians
    rollAngle = pa_scx

!*** replaced for time dependent boresight **** 03.02.2012 ATI ****
!    boresight = Cal_getBoresight(0.0_double);
!    boremat => CAL_toDirectionCosineMatrix(boresight) 
!*** 

    !***********************************************************
    ! Get the number of seconds since the mission reference time
    !***********************************************************
    time =  MissionTime(dataSet)

    !***********************************************************
    ! Get the  time dependent boresight  **** 03.02.2012 ATI
    !***********************************************************

    call CAL_GetBoresightMatrix(time, boremat) ! boresight offset matrix
    
    !*******************************************************************
    ! need inverted boresight matrix which is obtained via the transpose. 
    !********************************************************************
    inv_boremat = TRANSPOSE(boremat)

    attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle)

    !*******************************************************
    ! This routine requires a matrix of size 0:2, 0:2, but
    ! after that we want it to be of size 1:3, 1:3
    !******************************************************
    call OAL_toAttitudeMatrix(attitude, attmatrix1)
    
    forall( i = 1 : 3) 
        forall(j = 1 : 3) attmatrix(i, j) = attmatrix1(i - 1, j - 1)
    end forall

   
 
    !******************************************************************************
    ! converts RA and DEC from spherical cordinates to cartesian coordinates vector
    !*****************************************************************************
    call sla_dcs2c(ra_scx, dec_scx, pointvec)

    !********************************************************
    ! normalize the vector pointvec and return in pointvec2 -
    ! r1 holds the mdulus of the vector. pointvec2 will store
    ! the unit-vector in the direction of pointvec
    !********************************************************
    call sla_dvn(pointvec, pointvec2, r1)

    !
    !
    ! use SIMON'S BIT
    !
    !write(*,*)'using SIMONS BIT (essentially redundant now but keep for now
    !-----------------------------------------
    pointvec(1) = -pointvec2(1)
    pointvec(2) = -pointvec2(2)
    pointvec(3) =  pointvec2(3)
    !-----------------------------------------

 

    ! convert euler axis vector to a rotation matrix
     

    ! initialise output vector
    omvec = 0.0_double


!**********************************************************************************************
!  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
!  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
!  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
!
!  This creates omvec, the om boresight vector measured in the ERF
!  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
!  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
!  is the OM boresight as measured in the OM reference frame
!*********************************************************************************************

    att_invbor = matmul(attmatrix, inv_boremat)
    omvec = matmul(att_invbor, ompoint)

    !************************************************************************
    ! converts the omvec cartesian vector to ra and dec spherical coordinates
    ! and check RA lies within sensible range
    ! ra_point and ra_dec refer to the pointing direction of the instrument
    !************************************************************************

    call sla_dcc2s(omvec, ra_point, dec_point)

    if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi
 

    write(errstr, '(a, f8.4)') "RA of OM boresight: ", (ra_point * radiansToDegrees)
    call message(AppMsg, Verbose)
    write(errstr, '(a, f8.4)') "DEC of OM boresight: ", (dec_point * radiansToDegrees)
    call message(AppMsg, Verbose)
   

end subroutine OmPointingDirection


!****************************************************************
!The grism roll angle must be subtracted from the 
! rollAngle to get the real value of this parameter 
!
!***************************************************************
subroutine OmGrismPointingDirection(dataSet, ra_point, dec_point,grismRollAngle)
    type(dataSetT), intent(in) :: dataSet
    real(double), intent(in) :: grismRollangle             ! the grism filter roll angle     
    real(double), intent(out) :: ra_point, dec_point  ! position of the OM tangent point
    real(double) :: rollangle                  ! the roll angle
   
    real(double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
    real(double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

    real(double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
    real(double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
    real(double) :: pointvec(3)        ! s/c pointing
    real(double) :: pointvec2(3)       
    real(double), pointer :: boremat(:,:)       ! boresight offset matrix
    real(double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
    real(double) :: omvec(3)                   ! om pointing vector
  

    real(double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     
    real(double) :: r1                         ! modulus
    type (SpacecraftAttitudeType) attitude
     
    type(BoresightT) :: boresight    
    integer i, j    
    real(double) :: time

    ra_point = 0.0_double
    dec_point = 0.0_double


    !*******************************************************
    ! get the astrometry info keywords from the source list
    !*******************************************************
    ra_scx = real64Attribute(dataSet, "RA_SCX") * degreesToRadians   ! Spacecraft RA in radians
    dec_scx = real64Attribute(dataSet, "DEC_SCX") * degreesToRadians ! Spacecraft dec in radians
    pa_scx = real64Attribute(dataSet, "PA")        ! Spacecraft PA in degrees
 
    pa_scx = (pa_scx -grismRollAngle)* degreesToRadians
    rollAngle = pa_scx

!*** replaced for time dependent boresight **** 03.02.2012 ATI ****
!    boresight = Cal_getBoresight(0.0_double);
!    boremat => CAL_toDirectionCosineMatrix(boresight) 
!***

    !***********************************************************
    ! Get the number of seconds since the mission reference time
    !***********************************************************
    time =  MissionTime(dataSet)

    !***********************************************************
    ! Get the  time dependent boresight  **** 03.02.2012 ATI
    !***********************************************************

    call CAL_GetBoresightMatrix(time, boremat) ! boresight offset matrix

    
    !*******************************************************************
    ! need inverted boresight matrix which is obtained via the transpose. 
    !********************************************************************
    inv_boremat = TRANSPOSE(boremat)

    attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle)

    !*******************************************************
    ! This routine requires a matrix of size 0:2, 0:2, but
    ! after that we want it to be of size 1:3, 1:3
    !******************************************************
    call OAL_toAttitudeMatrix(attitude, attmatrix1)
    
    forall( i = 1 : 3) 
        forall(j = 1 : 3) attmatrix(i, j) = attmatrix1(i - 1, j - 1)
    end forall

   
 
    !******************************************************************************
    ! converts RA and DEC from spherical cordinates to cartesian coordinates vector
    !*****************************************************************************
    call sla_dcs2c(ra_scx, dec_scx, pointvec)

    !********************************************************
    ! normalize the vector pointvec and return in pointvec2 -
    ! r1 holds the modulus of the vector. pointvec2 will store
    ! the unit-vector in the direction of pointvec
    !********************************************************
    call sla_dvn(pointvec, pointvec2, r1)

    !
    !
    ! use SIMON'S BIT
    !
    !write(*,*)'using SIMONS BIT (essentially redundant now but keep for now
    !-------------------------------------------
    pointvec(1) = -pointvec2(1)
    pointvec(2) = -pointvec2(2)
    pointvec(3) =  pointvec2(3)
    !--------------------------------------------

 

    ! convert euler axis vector to a rotation matrix
     

    ! initialise output vector
    omvec = 0.0_double


!**********************************************************************************************
!  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
!  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
!  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
!
!  This creates omvec, the om boresight vector measured in the ERF
!  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
!  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
!  is the OM boresight as measured in the OM reference frame
!*********************************************************************************************

    att_invbor = matmul(attmatrix, inv_boremat)
    omvec = matmul(att_invbor, ompoint)

    !************************************************************************
    ! converts the omvec cartesian vector to ra and dec spherical coordinates
    ! and check RA lies within sensible range
    ! ra_point and ra_dec refer to the pointing direction of the instrument
    !************************************************************************

    call sla_dcc2s(omvec, ra_point, dec_point)

    if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi
 

    write(errstr, '(a, f8.4)') "RA of OM boresight:  ", (ra_point * radiansToDegrees)
    call message(AppMsg, Verbose)
    write(errstr, '(a, f8.4)') "DEC of OM boresight: ", (dec_point * radiansToDegrees)
    call message(AppMsg, Verbose)
   

end subroutine OmGrismPointingDirection



!******************************************************************
!
!*****************************************************************
subroutine CorrectTheAstrometry(nrows,  rightAscension, declination, &
            posErr, magnitude, ra_point, dec_point, dataSeth, &
            sourceListTable, corrData, corrWasOkay)     

    integer(int32), intent(in) :: nrows
    real(double), pointer ::  rightAscension(:), declination(:)
    real(single), pointer ::   posErr(:), magnitude(:)
    real(double), intent(in) :: ra_point, dec_point
    type(dataSetT), intent(inout) :: dataSetH
    type(TableT), intent(inout) :: sourceListTable
    logical, intent(out) :: corrWasOkay
    logical :: status = .false.
    real(single) :: maxRADecErr = 1.0, maxRmsResidual = 1.5, tolerance
    type(CorrectionData), intent(out) :: corrData
    type(CatalogueData), pointer :: catData(:) => null()
    real(double), pointer ::  rightAscension1(:) => null(), declination1(:) => null()
!    type(dataSetT) :: dataSetH
!    type(TableT) :: sourceTable
    logical :: okay = .false.
    !integer :: ii

    ! the variable for test 14.02.08 vny
    integer(int32) :: position
    real(double), dimension(:), pointer :: raCorr => null()
    real(double), dimension(:), pointer :: decCorr => null()

    corrWasOkay = .false. 


    tolerance = realParameter("tolerance")
    call setTolerance( tolerance )



    !**************************************************************
    ! Keep the old ras and decs in case the fit is no good
    !*************************************************************
    rightAscension1 => allocateDouble(1, size(rightAscension), "rightAscension1")
    declination1 => allocateDouble(1, size(rightAscension), "declination1")
    rightAscension1 = rightAscension * radiansToDegrees
    declination1 = declination * radiansToDegrees


    call CorrectAstrometry(nrows, rightAscension, declination, &
            posErr, magnitude, ra_point, dec_point,   dataSetH, &
            status, corrData, catData)     


 ! write(*, *) "Line 1259 corrwsokay=", corrwasOkay, " okay=", okay

 
 ! write(*,*) "L1262 status=", status

    if (status) then  ! The astrometry correction was successful
!        corrData%sigXShift =  sqrt(corrData%sigXShift *  corrData%sigXShift + 0.25)
!        corrData%sigYShift =  sqrt(corrData%sigYShift *  corrData%sigYShift + 0.25)

       maxRADEcErr = realParameter("maxradecerr")
       maxRmsResidual = realParameter("maxrmsres")

        ! Check if fitted parameters are okay 
        if (fitOkay(corrData, maxRADecErr, maxRmsResidual)) then
            rightAscension  = rightAscension * radiansToDegrees
            decLination = declination * radiansToDegrees
            okay = .true.
            call AddCorrectionColumns2(sourceListTable, nrows, rightAscension, decLination, corrData)

            ! Testing the results 14.02.2008 vny ===
            position = columnNumber(sourceListTable, "DEC_CORR")

            raCorr => AllocateDouble(1, nrows, "raCorr")
            decCorr => AllocateDouble(1, nrows, "decCorr")
            if (hasColumn(sourceListTable, "RA_CORR")) then
               call ReadReal64Column(sourceListTable, "RA_CORR", nRows, 0, raCorr)      
            else 
               raCorr = 0.0
            endif
            if (hasColumn(sourceListTable, "DEC_CORR")) then
               call ReadReal64Column(sourceListTable, "DEC_CORR", nRows, 0, decCorr)      
            else 
               decCorr = 0.0
            endif
           
            ! ===
        else
            call AddCorrectionColumns2(sourceListTable, nrows, rightAscension1, decLination1, corrData)
   
        end if    
       
    else 
        
        call AddCorrectionColumns2(sourceListTable, nrows, rightAscension1, decLination1, corrData)
   
        if (associated(catData)) then
        
            call pgbegin(0, "pgplot.ps/VC", 1, 1)


          
            call DrawGraph(nrows, real(rightAscension), &
                  real(decLination), "omatt", "RA", "DEC", 0.1, 0.9, 0.1, 0.9)
            call pgsci(3)
            call pgpoint(size(catData), real(catData%Ra * radiansToDegrees), &
                  real(catData%dec * radiansToDegrees), 3)
            call pgclos
        end if
       
   
    end if

    deallocate(rightAscension1, declination1)

!    dataSetH = dataSet(sourceListSet, MODIFY)
!    sourceTable = table(dataSetH, "SRCLIST")
   
    corrWasOkay = okay
   

!write(*, *) "Line 1321 corrwsokay=", corrwasOkay, " okay=", okay
    call setAttribute(sourceListTable, "POSCOROK", okay, &
                              "ASTROMETRY CORRECTION OKAY")
    call setAttribute(sourceListTable, "SYSERROR", 0.5, &
                              "Systematic catalogue position errors")

!    call release(sourceTable)
!    call release(dataSetH)
end subroutine CorrectTheAstrometry

!*******************************************
! Function ComputeEpoch
!******************************************
real(double) function  ComputeEpoch(dateObs)
   
    character(19), intent(in)  :: dateObs               
    real(double)  :: JulianDate0, JulianDate
    character(19)  :: missionReferenceTime = "1998-01-01T00:00:00"
    
    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the observation
    JulianDate = ModifiedJulianDate(dateObs)

    
    ComputeEpoch = 1998.0_double + (JulianDate - JulianDate0) / 365.25_double

end function ComputeEpoch

!*******************************************************************
!
!
!******************************************************************
subroutine ComputeGrismSkyCoordinates(xPos, yPos, x0, y0, &
                                       xBinning, yBinning, &
                                       plateScale, rollAngle, grismRollAngle, &  
                                       ra_point, dec_point, &
                                       rightAscension, declination , &
                                       galacticLong, galacticLat)
    real(single), intent(in) :: xPos, yPos
    real(single), intent(in) :: x0, y0
 
    real(single), intent(in) :: xBinning, yBinning
    real(double), intent(in) :: plateScale, rollAngle, grismRollAngle
    real(double), intent(in) :: ra_point, dec_point
  
    real(double), intent(out) :: rightAscension, declination
    real(double), intent(out) :: galacticLong, galacticLat
    real(double) ::  xposcorr,  yposcorr
    real(double) :: xi_source, eta_source
    real(double) :: xrad, yrad              ! position in radians     
    real(double) :: cosRollAngle, sinRollAngle
    real(double) :: cosGRollAngle, sinGRollAngle ! cos and sin of the grism roll angle
    integer(int32) :: mode


    cosRollAngle = dcos(rollAngle)
    sinRollAngle = dsin(rollAngle)
    cosGRollAngle = dcos(grismRollAngle)
    sinGRollAngle = dsin(grismRollAngle)     

    xposcorr = dble(xpos) * dble(xBinning) + dble(x0)  
    yposcorr = dble(ypos) * dble(ybinning) + dble(y0)
 

    !***************************************************
    ! correct positions for pincushion/barrel distortion
    !***************************************************
  
    rawx(0) = real(xposcorr)
    rawy(0) = real(yposcorr)


    mode = 1
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, mode)


    xposcorr = xposcorr - dble(deltax(0))  ! - sign removes distortion    
    yposcorr = yposcorr - dble(deltay(0))  ! - sign removes distortion

           
    !***************************************************
    ! get offset relative to tangent point (at OM centre)
    !***************************************************
    xrad = xposcorr - xCentre 
    yrad = yposcorr - yCentre


    !**************************************************
    !  rotate x,y frame to tangent sky (RA, DEC) frame
    !  and form tangent-plane coords
    !*************************************************
    xi_source = -plateScale * (xrad * cosRollAngle - yrad * sinRollAngle)
    eta_source = plateScale * (xrad * sinRollAngle + yrad * cosRollAngle)


    !***********************************************************
    ! Transform tangent-plane coordinates (xi_source, yi_source) 
    ! into spherical (sky) coordinates (RA, dec). ra_point,
    ! dec_point is the spherical coordinate of the tangent-point
    ! Also calculate the galactic longitude and latitude of
    ! each source
    !************************************************************
     
    call sla_dtp2s(xi_source, eta_source, &
            ra_point, dec_point, rightAscension, declination)

    call sla_EQGAL(rightAscension,  declination, &
                      galacticLong, galacticLat)


end subroutine ComputeGrismSkyCoordinates

!subroutine invert_trans(trans, itrans)
!    real(kind=double), intent(in) ::  trans(6)
!    real(kind=double), intent(out) :: itrans(6)
!    real(kind=double) determ!
!
!    determ = trans(2) * trans(6) - trans(3) * trans(5)
!    itrans(1) = trans(3) * trans(4) - trans(1) * trans(6)
!    itrans(2) = trans(6)
!    itrans(3) = -trans(3)
!    itrans(4) = trans(1) * trans(5) - trans(2) * trans(4)
!    itrans(5) = -trans(5)
!    itrans(6) = trans(2)
!    itrans = itrans / determ
!
!end subroutine invert_trans

!*******************************************************************
!  ComputeGrismSkyCoordinates2 : modified for unrotating 
! the source x and y positions
!******************************************************************
subroutine ComputeGrismCoordinates2(isrc, xPos, yPos, x0, y0, &
     xBinning, yBinning, plateScale, rollAngleIn, grismRollAngle, &  
     ra_point, dec_point, rightAscension, declination , &
     galacticLong, galacticLat,omgXmin,omgYmin, windowX0, &
     windowY0, grismOffsetX, grismOffsetY)
!    grismOffsetX and grismOffsetY are in arcsecond in the input
    integer(int32), intent(in) :: isrc
    integer(int32) :: isrc2
    real(single), intent(in) :: xPos, yPos
    real(single), intent(in) :: x0, y0
    real(single) :: x02, y02

    real(single), intent(in) :: xBinning, yBinning
    real(double), intent(in) :: plateScale, rollAngleIn, grismRollAngle
    real(double), intent(in) :: ra_point, dec_point
  
    real(double), intent(out) :: rightAscension, declination
    real(double), intent(out) :: galacticLong, galacticLat
    real(double), intent(in) :: omgXmin, omgYmin
    real(double), intent(in) :: grismOffsetY, grismOffsetX
    integer(kind=int32) :: windowX0, windowY0
    real(double) ::  xposcorr,  yposcorr
    real(double) :: xi_source, eta_source
    real(double) :: xrad, yrad              ! position in radians     
    real(double) :: cosRollAngle, sinRollAngle
    real(double) :: cosGRollAngle, sinGRollAngle ! cos and sin of the grism roll angle
    integer(int32) :: mode

    real(single) :: xPos2, yPos2

    real(kind=double)  :: trans(6), itrans(6)
    real(double) :: grismPlateScaleY, grismOffsetYR, grismOffsetXR
    !real(double) :: grismPlateScaleY, grismPlateScaleX, grismOffsetYR, grismOffsetXR
    real(double) :: rollAngle

    isrc2=isrc  ! dummy assignment to avoid a warning message
    x02=x0
    y02=y0

    grismPlateScaleY=0.52_double ! this is computed by using USNO RA and DECs of the sources
    !grismPlateScaleX=0.49_double
    grismPlateScaleY=degreesToRadians*grismPlateScaleY/3600.0_double
    !grismPlateScaleX=degreesToRadians*grismPlateScaleX/3600.0_double

    !grismOffsetY=-9.72_double
    !grismOffsetX=0.0_double
    grismOffsetYR=degreesToRadians*grismOffsetY/3600.0_double   
    grismOffsetXR=degreesToRadians*grismOffsetX/3600.0_double 

    !if(x0 == y0 .and. grismPlateScaleX==grismPlateScaleY .and. isrc == 0) then
    !end if



    xPos2=xPos
    yPos2=yPos

! VNY 3.11.2006 rollAngle outside of this routine was corrected to build the sky-image
!               so we have to return it to its normal value
!               (the correction was 2.0*grismRollAngle) but we return only the single value
    rollAngle=rollAngleIn+grismRollAngle*degreesToRadians

    cosRollAngle = dcos(rollAngle)
    sinRollAngle = dsin(rollAngle)
    
    !grismRollAngle=grismRollAngle*degreesToRadians
    cosGRollAngle = dcos(grismRollAngle*degreesToRadians)
    sinGRollAngle = dsin(grismRollAngle*degreesToRadians)     
    
    trans(1) =  -omgXmin
    trans(2) =  dcos(-grismRollAngle*degreesToRadians)
    trans(3) =  dsin(-grismRollAngle*degreesToRadians)
    trans(4) =  -omgYmin
    trans(5) = -dsin(-grismRollAngle*degreesToRadians)
    trans(6) =  dcos(-grismRollAngle*degreesToRadians)

    call invert_trans(trans, itrans)
 
    xposcorr = dble(xpos)  
    yposcorr = dble(ypos)

!*********************************************************
! For grisms before correcting for distortion we have to 
! unrotate the source position (because we are using the 
! image with the spectrum aligned vertically)
!**********************************************************
!ftp://xsa.esac.esa.int:21/pub/AIOURL494075/0092820801/odf/0092820801.tar.gz

    !write(8,*) "L1199 xCentre=", xCentre, " yCentre=", yCentre, " x0=", x0, " y0=", y0

    !***************************************************
    ! get offset relative to tangent point (at OM centre)
    !***************************************************
!    xrad = xposcorr - xCentre 
!    yrad = yposcorr - yCentre

    xrad = xposcorr 
    yrad = yposcorr
 
    !**************************************************
    !  rotate x,y frame to tangent sky (RA, DEC) frame
    !  and form tangent-plane coords
    !*************************************************
!    xi_source = -plateScale * (xrad * cosGRollAngle - yrad * sinGRollAngle)
!    eta_source = plateScale * (xrad * sinGRollAngle + yrad * cosGRollAngle)
!
!    xi_source =  -xrad * cosGRollAngle + yrad * sinGRollAngle
!    eta_source = xrad * sinGRollAngle + yrad * cosGRollAngle
!
  !+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
  ! Rotate to get the initial window position (before omgprep -
  ! when the spectra were not aligned with the y-axis
  !+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
    xi_source =  itrans(1) + xrad *itrans(2) + yrad * itrans(3)
    eta_source = itrans(4) + xrad *itrans(5) + yrad * itrans(6)

    xi_source=xi_source+windowX0
    eta_source=eta_source+windowY0

    !***************************************************
    ! correct positions for pincushion/barrel distortion
    !***************************************************
  
    rawx(0)=real(xi_source)
    rawy(0)=real(eta_source)

    mode = 1
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, mode)

    xposcorr = rawx(0) - dble(deltax(0))  ! - sign removes distortion    
    yposcorr = rawy(0) - dble(deltay(0))  ! - sign removes distortion

           
    !***************************************************
    ! get offset relative to tangent point (at OM centre)
    !***************************************************
    ! xCentre and yCentre must be corrected for the binning factor
    xrad = xposcorr - xCentre/xbinning 
    yrad = yposcorr - yCentre/ybinning
    ! Here, in order to convert into RA and DECwe have to use the initial roll-angle 

!**************************************************
    !  rotate x,y frame to tangent sky (RA, DEC) frame
    !  and form tangent-plane coords
    !*************************************************
    xrad = plateScale * xrad*xbinning
    yrad = grismPlateScaleY * yrad*ybinning

    xi_source = xrad * cosRollAngle - yrad * sinRollAngle
    eta_source = xrad * sinRollAngle + yrad * cosRollAngle

    xi_source=xi_source+grismOffsetXR
    eta_source=eta_source+grismOffsetYR

    xi_source=-xi_source

    !***********************************************************
    ! Transform tangent-plane coordinates (xi_source, yi_source) 
    ! into spherical (sky) coordinates (RA, dec). ra_point,
    ! dec_point is the spherical coordinate of the tangent-point
    ! Also calculate the galactic longitude and latitude of
    ! each source
    !************************************************************
     
    call sla_dtp2s(xi_source, eta_source, &
            ra_point, dec_point, rightAscension, declination)
    call sla_EQGAL(rightAscension,  declination, &
                      galacticLong, galacticLat)

end subroutine ComputeGrismCoordinates2


!****************************************************************
! Subroutine GetGrismPositoon (obtained by transforming the 
! subroutine GetTargetPosition) to get the X and Y coordinates 
! of the grism image zero-point (in arcseconds) for each filter
! and resolution
!****************************************************************
subroutine GetGrismPosition(primInstrValue, filter, binBpe, xbin, &
     ybin, grismOffsetX, grismOffsetY)
    character(len=*), intent(in) :: primInstrValue, filter
    logical(kind=Bool), intent(in) :: binBpe
    integer, intent(in) :: xbin, ybin
    real(double), intent(out) :: grismOffsetX, grismOffsetY
    real(double), dimension(3,2,2) :: xCal, yCal
    integer :: i,j,k

! The offsets adgusted by hand
    data xCal/-0.7_double, -5.1_double, 7.0_double, 2.07_double, &
         -4.2_double, 7.01_double, 156.06_double, 48.8_double,   &
         -90.2_double, 250.07_double, -20.1_double,-20.6_double/
    data yCal/-6.4_double, 9.3_double,  9.24_double, -7.5_double, 8.28_double, &
              10.9_double, -303.5_double, 337.2_double,-320.1_double,&
               -303.9_double,-390.0_double,-390.5_double/
!==================================================================
!  This is a modified table for the grism offsets (in arcseconds)
!-------------------------------------------------------------------
!        R1     !       PN (E3)   !        OM         ! <-- Prim.instr
!-------------------------------------------------------------------
!     X     Y   !    X       Y    !     X      Y      !
!-------------------------------------------------------------------
! -0.7*  -6.4*  !   -5.1*   9.3*  !  7.0*    9.24*    ! <-- UV, HR
! 2.07*  -7.5*  !   -4.2*  8.28*  !  7.01*   10.9*    ! <--  V, HR
!               !                 !                   ! 
!156.06* -303.5*!   48.8*  337.2* !  -90.2** -320.1** ! <-- UV, LR
! (22.7* 386.0*)!  -20.1*   -390* !  -20.6   -390.5   ! <--  V, LR
!250.07* -303.9*! <- using this, derived from 0655_0159960101 (V,LR)
!------------------------------------------------------------------- 
!The values labelled (*) are checked
!(**) currently there are no observations with LR,UV or V for Prime=OM
!
! There is an ambiguity for R1,V,LR:
! the data set 0655_0159960101 shows the position (+250.07,-303.9) arcseconds,
! whereas the set  0462_0125910201 gave (+22.7,+386.0) arcseconds.
! The latest data set is taken for the calibration  
!    
!    i=3 ! default primary instrument: OM (if no primary found)
    i=1 ! default primary instrument: R1 (if no primary found)
    if (trim(primInstrValue) .eq. "R1") i=1
    if (trim(primInstrValue) .eq. "OM") i=3
    if (trim(primInstrValue) .eq. "E3") i=2
    if (trim(primInstrValue) .eq. "PN") i=2
!  
    
    j=1 ! filter=UV (grism1)
    if (trim(filter) .eq. "GRISM2" .or. trim(filter) .eq. "Grism2") j=2

    k=1 ! resolution is High (binBpe=F)
    if (binBpe) k=2
    grismOffsetX=xCal(i,j,k)
    grismOffsetY=yCal(i,j,k)
    if (xbin .gt. 1 .and. .not. binBpe) grismOffsetX=grismOffsetX/xbin
    if (ybin .gt. 1 .and. .not. binBpe) grismOffsetY=grismOffsetY/ybin
    !write(*,*) "L1560 xbin=", xbin, " primInstr=", trim(primInstrValue), &
    !     " filter=", trim(filter), " dx=", grismOffsetX, &
    !     " dy=", grismOffsetY
    write(errstr, '(a,a,a,a,a,i2)') "Prime instrument: ", trim(primInstrValue), &
         " Filter=", trim(filter), " Binning factor=", xbin
    call message(AppMsg, Verbose)

end subroutine GetGrismPosition

!***************************************************************
! Subroutine GetPrimaryInstrument
!
! This subroutine will find the primary instrument name using
! the summary-file
!****************************************************************
subroutine  GetPrimaryInstrument(PPSOSWSet,primInstrValue) 
  character(len=255), intent(in) :: PPSOSWSet   
  character(len=10), intent(out) :: primInstrValue
  integer :: nExps, count1
  character(len=1024)  :: summaryFileName, summaryShortFileName, outDirectory
  character(len=1024) :: modifiedSummaryFileName, summaryString
  integer(kind=int16) :: posPoint, primPos, pos1, pos2
  logical :: existFile, eofReached, primInstrFound, existModifiedFile
  logical :: existWithinOutDirectory
       
  nExps = OAL_selectScope(XMM, SUMMARY_INFORMATION)   

  call OAL_selectFile(0, summaryFileName);
  posPoint=0
  posPoint=index(trim(summaryFileName),'/',.true.) ! .true. means backward
  summaryShortFileName=summaryFileName 
  if (posPoint .gt. 3) then
     summaryShortFileName(1:posPoint)=' '
     summaryShortFileName=adjustl(summaryShortFileName)
  endif

  ! Find the name of the output directory
  posPoint=0        
  outDirectory=PPSOSWSet
  posPoint=index(trim(outDirectory),'/',.true.) ! .true. means backward
  if (posPoint .ge. 1) then
     outDirectory(posPoint:1024)=' '
     ! combine the name of the summary file and the output directory
     outDirectory(posPoint:posPoint+1)='/ '
     outDirectory(posPoint+1:1024)=summaryShortFileName(1:1024-posPoint)
  else
     outDirectory='./'
     outDirectory(3:1024)=summaryShortFileName(1:1021)
  end if
  existFile=.false.
  existModifiedFile=.false.
  existWithinOutDirectory=.false.
  primPos=0
  primPos=index(trim(summaryFileName),'SUM.')
  ! Find summary file in the usual directory (ODF)
  inquire (FILE=trim(summaryFileName), EXIST=existFile)

  if (.not. existFile) then
     ! Search for the original (ascii) summary file 
     modifiedSummaryFileName=trim(summaryFileName)
     if (modifiedSummaryFileName(primPos+4:primPos+6) .eq. 'SAS') then
        modifiedSummaryFileName(primPos+4:primPos+6)='ASC'
     else
        modifiedSummaryFileName(primPos+4:primPos+6)='SAS'
     endif
     inquire(FILE=trim(modifiedSummaryFileName), EXIST=existModifiedFile)
     if (.not. existModifiedFile) then
        inquire(FILE=trim(outDirectory),EXIST=existWithinOutDirectory)
     endif
  endif

  if (existFile .or. existModifiedFile .or. existWithinOutDirectory) then
     ! Open the summary file
     if (existFile) open(2,file=trim(summaryFileName))
     if (existModifiedFile) open(2,file=trim(modifiedSummaryFileName))
     if (existWithinOutDirectory) open(2,file=trim(outDirectory))
     eofReached=.false.
     primInstrFound=.false.
     count1=0
     primPos=0
     primInstrValue=''
     do
        read(2, '(a)', end=110) summaryString
        count1=count1+1
        primPos=index(summaryString,'PRIME_INSTR')
        if (primPos .gt. 0) then
           primInstrFound=.true.
           exit
        endif
     enddo
110  continue
120  close(2)         
     if (primInstrFound) then
        pos1=index(summaryString, '=')
        pos2=index(summaryString, '/')
        primInstrValue=summaryString(pos1+2:pos2-1)
     else
        primInstrValue=''
        call warning("SummaryFile", "Primary instrument string not found in the summary file")
     endif
  else
     primInstrValue=''
     call warning("SummaryFile", "No summary file found")
  endif

end subroutine GetPrimaryInstrument

end module omatt_mod

