!****************************************************************************************
!
! Given an input image file, rotates it by the given angle
! and then save it to a fits file with the given name
!
!*****************************************************************************************  
! 23.02.2006 VNY slightly correcting the code in order to disable undistortion 
! when processing GRISM images
! 08.11.2006 VNY Corrections needed for the grism images astrometry
! 07.04.2008 VNY added two new attributes to the rotated (sky) image 
!                FITs file: RA_OFF and DEC_OFF (in arcseconds) to indicate 
!                the amount of astrometric correction applied to the image
!          
module omrotation_mod

  use oal
  use cal
  use dal
  use sasdal
  use caloaldefs

  use errorHandling
  use sla
  use readModes_mod  
  use omallocation_mod 


  use omconstants_mod
  use omastromcorrect_mod    ! For doing the astrometry correctionr
  use omdistortion_mod
  use OmSourceparam_mod
implicit none


public
    
   real(double) :: xOrigin, yOrigin  
   real(double) :: xOrigin1, yOrigin1  
    
   real(double) :: dblNx, dblNy
   real(double)  :: sinRollAngle, cosRollAngle
  
   real(single), parameter :: stepX = 0.5  
   real(single), pointer :: interpolationImage(:, :) => null()

   real(double) :: xBin, yBin 
  
    
   type OmImage
       integer(int16) :: x0      ! window x0
       integer(int16) :: y0      ! window y0
       integer(int16) :: binax1  ! x binning factor
       integer(int16) :: binax2  ! y binning factor
       integer(int16) :: crpix1  ! pixel i of reference point
       integer(int16) :: crpix2  ! pixel j of reference point
       real(single)   :: crval1  ! ra at reference point
       real(single)   :: crval2  ! dec at reference point
      
       real(single)   :: cdelt1  ! arcsecs/pixel along x
       real(single)   :: cdelt2  ! arcsecs/pixel along y
   end type OmImage
contains


!************************************************
! CreateRotatedImage
!
!  Creates THE OUTPUT PPS PRODUCT OM OSW SKY FITS IMAGE. 
!  This is done by resampling the detector image on a sky grid. 
! the transformation is currently a simple rotation
! Define the transformation terms (note the offsets will change)
! 15.02.2008: the RA and DEC shifts as determined by the astrometry
! correction (if the parameter usecat were set to "yes") are
! applied to the CRVAL1 and CRVAL2 attributes of the image
!****************************************************************
subroutine CreateRotatedImage(set, PPSOSWSET, rollAngle, ra_point, dec_point, &
              xBinning, yBinning, ra_scx, dec_scx, plateScale, posCorr, ra_off, dec_off)
    character(*), intent(in) :: set                                         ! OM image intermediate filename
 
    character(*), intent(in) :: PPSOSWSet    ! PPS OM OSW FITS sky image filename.

    real(double), intent(in) :: rollAngle
    real(double), intent(in) :: ra_point, dec_point        ! position of the OM tangent point

    integer, intent(in)        :: xBinning
    integer, intent(in)        :: yBinning
    real(double), intent(in) :: ra_scx, dec_scx  ! The ra, dec and pa of the spacecraft     
   
    real(double), intent(in) :: plateScale         ! SRR added 4/10/00
    !real(double), intent(in) :: plateScaleAux
 
    real(single) :: refx, refy
  

    type(DataSetT) :: inDataSet                                       ! Input dataset handle.

    type(ArrayT) :: inArrayH                                          ! Input array handle.
 
    type(DataSetT) :: PPSOSWDataSetH                                  ! PPS OSW dataset handle.

    type(ArrayT) :: PPSOSWArrayH                                      ! PPS OSW array handle.


    real(double)  :: trans(6) ! rotation transformation matrix
    real(double)  :: itrans(6)     ! its inverse

    real(single), pointer :: oswImage(:, :) => null()
 
    real(single), pointer  :: outimage(:,:) => null()      ! The output image
    real(single)          :: exposureTime         ! The exposure time for an input image
  
    real(double)  :: sinRollAngle, cosRollAngle
    integer             :: idmsf(4), ihmsf(4)  
    integer             :: nxout, nyout        ! size of the output image
    integer                    :: dim(2), nx, ny     ! size of the image

    real(single)                       :: refDX
    real(single)                       :: refDY
    real(single)                       :: rotatedRefDX
    real(single)                       :: rotatedRefDY
    real(double)  :: xmax,xmin,ymax,ymin            
    integer(int32) ::  i, j, k    
    character(20):: rastring, decstring   ! strings for RA and DEC
    character(1) :: cs                    ! sign of dec +/-
    integer             :: x0, y0             ! offset of the image
    integer             :: xsize, ysize       ! size of the image

    character(6)            :: omFilter   ! the OM filter
    character(21) :: detsecValue
    character(1) :: auxc1
    character(6) :: auxc6
    integer :: detsecx0, detsecy0, rotateddsx0, rotateddsy0
    logical(bool) :: grismFlag

    real(double) :: crpix1, crpix2
    logical(bool), intent(in) :: posCorr  
    real(single), intent(in) :: ra_off, dec_off ! ra and dec corrections (arcsecs), which were applied
    !write(*,*) "L134 ra_off=", ra_off
    grismFlag=.false.

    !plateScaleAux=PlateScale

    xBin = xBinning
    yBin = yBinning
    sinRollAngle = dsin(rollAngle)  ! Need these quantities quite a lot
    cosRollAngle = dcos(rollAngle)
  
 
    inDataSet = DataSet(set, modify) 

    inArrayH = array(inDataSet, "PRIMARY", modify)

    omFilter = stringAttribute(inDataSet, "FILTER")
    if (trim(omFilter) .eq. "Grism1" .or. trim(omFilter) .eq. "GRISM1" &
        .or. trim(omFilter) .eq. "Grism2" .or. trim(omFilter) .eq. "GRISM2" ) then
       grismFlag=.true.
       crpix1=real64Attribute(inDataSet, "CRPIX1")
       crpix2=real64Attribute(inDataSet, "CRPIX2")
    endif

    ! in order to transform correctly the grism image into the sky coordinates
    ! we need the shift of the left lower corner recorded in DETSEC
    if (grismFlag) then
       detsecValue=stringAttribute(inDataSet,"DETSEC")
       read(detsecValue,'(a1,i4,a6,i4)') auxc1,detsecx0,auxc6,detsecy0
    else
       detsecx0=0
       detsecy0=0
    endif
    exposureTime = real32Attribute(inDataSet, "EXPOSURE") ! Image exposure time

    !***********************************************************
    ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
    !***********************************************************
    call readModes(inDataSet, x0, y0, xsize, ysize)  
    oswImage => real32array2data(inArrayH)
 
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)
    trans(1) =  0.0_double
    trans(2) =  dcos(-rollangle)
    trans(3) =  dsin(-rollangle)
    trans(4) =  0.0_double
    trans(5) = -dsin(-rollangle)
    trans(6) =  dcos(-rollangle)

    !*******************************************
    ! create the output source file
    !*******************************************        
 
    PPSOSWDataSetH = DataSet(PPSOSWSet, create)

    !**********************************************************************
    ! Determine the bounding x and y (xmin, xmax, ymin, ymax) coordinates 
    ! of the image
    !*********************************************************************
    call  GetImageBoundaries(nx, ny, real(x0), real(y0), xBinning, yBinning, &
                             trans, xMin, xMax, yMin, yMax, grismFlag)

! invert transformation matrix for later use
    call invert_trans(trans, itrans)    
! need inverted transformation matrix later to rotate image

    nxout = xmax - xmin                  ! determine image dimensions
    nyout = ymax - ymin


!********************************************************
!
! POSSIBLE BUG xmax & xmin are boundaries of the image
! in double precision
! nxout should be xmax - xmin + 1
!********************************************************

 
!***********************************************************************************
! set up the output image with the appropriate size (xmax,ymax already referenced to 
! origin of containing (output) box so these equal the output box dimensions).
!**********************************************************************************

    dim(1) = nxout     ! The dimensions of the output image
    dim(2) = nyout
    PPSOSWArrayH  = addArray(PPSOSWDataSetH, "PRIMARY",  Real32, dim)  ! For the image array


    ! copy the header across

    call copyAllAttributes(attributable(inDataSet), attributable(PPSOSWDataSetH))
   

    ! add content keyword

    call setAttribute(PPSOSWDataSetH, "CONTENT", "OM OSW FITS SKY IMAGE", &
         "Content of FITS file")

!********************************************
! convert from ra and dec numbers to strings
!********************************************

    call SLA_CR2TF(2, real(ra_point), cs, ihmsf)
    write(rastring, '(i2.2,'':'',i2.2,'':'',i2.2,''.'',i2.2)') &
         ihmsf(1), ihmsf(2), ihmsf(3), ihmsf(4)
    call SLA_CR2AF(1, real(dec_point), cs, idmsf)
    write(decstring, '(a1,i2.2,'':'',i2.2,'':'',i2.2,''.'',i1)') &
         cs, idmsf(1), idmsf(2), idmsf(3), idmsf(4)

!******************************************************************************
! now determine refDX, refDY offsets of OM plate centre (PC) from original OSW 
! origin (IN BINNED PIXELS)
!******************************************************************************
    if (grismFlag) then
       !refDX = (xCentre - x0 ) / dble(Xbinning) - detsecx0  
       !refDY = (yCentre - y0 ) / dble(Ybinning) - detsecy0  
       !
       refDX=crpix1!-detsecx0
       refDY=crpix2!-detsecy0
    else
       refDX = (xCentre - x0) / dble(Xbinning)
       refDY = (yCentre - y0) / dble(Ybinning)
    endif

! Compute these offsets in terms of the rotated frame (remember rotation 
! takes place about the OM plate centre so the x,y offsets of the OSW origin 
! shift in the rotated frame). 

    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 

! Finally compute offsets (refx, refy) of the OM plate centre from the BB origin, i.e. 
! just add the offset of the PC from the OSW origin to the offset of the OSW origin 
! from the BB origin
    rotateddsx0 = detsecx0 * cosRollangle - detsecy0 * sinRollangle
    rotateddsy0 = detsecy0 * sinRollangle + detsecy0 * cosRollangle 

    refx = trans(1) + rotatedRefDX
    if (grismFlag) then
       !refx=refx+rotateddsx0!*Xbinning
       !refy=rotatedRefDY!-rotateddsy0!*Ybinning
       refy= trans(4) + rotatedRefDY
    else
       refy = trans(4) + rotatedRefDY
    endif

!
! And write the WCS attributes to output file. Note that the scale is in binned pixels (ie 
! if binning is 2, with raw scale -0.5" per pixels, CDELT1/2 =1
!

    !****************************************************
    ! Set the fits header attributes for the image array
    !****************************************************
    call SetFitsArrayAttributes(PPSOSWArrayH, xBinning, yBinning, plateScale, refx, refy, &
                                  ra_scx, dec_scx, ra_point, dec_point, grismFlag, posCorr, &
                                  ra_off, dec_off)


   
     


  








    outimage => real32array2data(PPSOSWArrayH)

    !*********************************************************  
    !  Now rotate the image.  
    !*********************************************************

    k = 0
    do i = 0, nxout - 1
        do j = 0, nyout - 1
            call setToNull( PPSOSWArrayH, k)

            k = k + 1
        end do
    end do

    ! Actually rotate the image (stored in outImage)
    ! VNY 23.02.2006 introduced the parameter grismFlag
    call RotateImage(nx, ny, nxout, nyout, oswImage, outImage, itrans, grismFlag)  

 
    ! Release handles

    call release(inArrayH)
    
    call release(PPSOSWArrayH) 
    
    call release(PPSOSWDataSetH)



    call release(inDataSet)

! Now check the header keywords CTYPE1 and CTYPE2 (only for the grim image)
    if (grismFlag) then
       PPSOSWDataSetH = DataSet(PPSOSWSet, modify)
       !PPSOSWArrayH  = array(PPSOSWDataSetH, "PRIMARY", modify) 

       if (hasAttribute(PPSOSWDataSetH, "CTYPE1")) then
          call deleteAttribute(PPSOSWDataSetH ,"CTYPE1")
          call setAttribute(PPSOSWDataSetH, "CTYPE1", "RA---TAN", "Coordinate system for X")
       endif!


       if (hasAttribute(PPSOSWDataSetH, "CTYPE2")) then
          call deleteAttribute(PPSOSWDataSetH,"CTYPE2")
          call setAttribute(PPSOSWDataSetH, "CTYPE2", "DEC--TAN", "Coordinate system for Y")
       endif
       call release(PPSOSWDataSetH)
    endif

end subroutine CreateRotatedImage



!************************************************
! CreateGrismSkyImage
! (derived from CreateRotatedImage)
!
!  Creates THE OUTPUT PPS PRODUCT OM OSW SKY FITS IMAGE. 
!  This is done by resampling the detector image on a sky grid. 
! the transformation is currently a simple rotation
! Define the transformation terms (note the offsets will change)
!****************************************************************
subroutine CreateGrismSkyImage(set, PPSOSWSET, rollAngle, grismRollAngle, &
     ra_point, dec_point, xBinning, yBinning, ra_scx, dec_scx, &
     plateScale, omgXref, omgYref)
    character(*), intent(in) :: set                                         ! OM image intermediate filename
 
    character(*), intent(in) :: PPSOSWSet    ! PPS OM OSW FITS sky image filename.

    real(double), intent(in) :: rollAngle
    real(double), intent(in) :: grismRollAngle ! (in degrees)
    real(double) :: modifiedRollAngle
    real(double), intent(in) :: ra_point, dec_point        ! position of the OM tangent point
    real(double), intent(in) :: omgXref, omgYref
    integer, intent(in)        :: xBinning
    integer, intent(in)        :: yBinning
    real(double), intent(in) :: ra_scx, dec_scx  ! The ra, dec and pa of the spacecraft     
   
    real(double), intent(in) :: plateScale         ! SRR added 4/10/00
 

    real(single) :: refx, refy
  

    type(DataSetT) :: inDataSet                                       ! Input dataset handle.

    type(ArrayT) :: inArrayH                                          ! Input array handle.
 
    type(DataSetT) :: PPSOSWDataSetH                                  ! PPS OSW dataset handle.

    type(ArrayT) :: PPSOSWArrayH                                      ! PPS OSW array handle.


    real(double)  :: trans(6) ! rotation transformation matrix
    real(double)  :: itrans(6)     ! its inverse

    real(single), pointer :: oswImage(:, :) => null()
 
    real(single), pointer  :: outimage(:,:) => null()      ! The output image
    real(single)          :: exposureTime         ! The exposure time for an input image
  
    real(double)  :: sinRollAngle, cosRollAngle
    real(double)  :: sinModifiedRollAngle, cosModifiedRollAngle
    integer             :: idmsf(4), ihmsf(4)  
    integer             :: nxout, nyout        ! size of the output image
    integer                    :: dim(2), nx, ny     ! size of the image

    real(double)                       :: refDX
    real(double)                       :: refDY
    real(double)                       :: rotatedRefDX
    real(double)                       :: rotatedRefDY
    real(double)  :: xmax,xmin,ymax,ymin  
    real(double)  :: xmaxr,xminr,ymaxr,yminr            
    integer(int32) ::  i, j, k    
    character(20):: rastring, decstring   ! strings for RA and DEC
    character(1) :: cs                    ! sign of dec +/-
    integer             :: x0, y0             ! offset of the image
    integer             :: xsize, ysize       ! size of the image

    character(6)            :: omFilter   ! the OM filter
    character(21) :: detsecValue
    character(1) :: auxc1
    character(6) :: auxc6
    integer :: detsecx0, detsecy0, rotateddsx0, rotateddsy0
    logical(bool) :: grismFlag

    real(double) :: crpix1, crpix2
   
    real(double) :: plateScaleCorrection, scaleAlpha, scaleDelta

    !--------------------------------------------------------------
    ! grism is increasing the scale along the dispersion direction
    !--------------------------------------------------------------
    plateScaleCorrection=0.0736_double
  

    grismFlag=.false.

    xBin = xBinning
    yBin = yBinning
    modifiedRollAngle=rollAngle-grismRollAngle*degreesToRadians
    sinRollAngle = dsin(rollAngle)  ! Need these quantities quite a lot
    cosRollAngle = dcos(rollAngle)
    !
    sinModifiedRollAngle = dsin(modifiedRollAngle)  ! Need these quantities quite a lot
    cosModifiedRollAngle = dcos(modifiedRollAngle)
  
    inDataSet = DataSet(set, modify) 

    inArrayH = array(inDataSet, "PRIMARY", modify)

    omFilter = stringAttribute(inDataSet, "FILTER")
    if (trim(omFilter) .eq. "Grism1" .or. trim(omFilter) .eq. "GRISM1" &
        .or. trim(omFilter) .eq. "Grism2" .or. trim(omFilter) .eq. "GRISM2" ) then
       grismFlag=.true.
       crpix1=real64Attribute(inDataSet, "CRPIX1")
       crpix2=real64Attribute(inDataSet, "CRPIX2")
    endif

    ! in order to transform correctly the grism image into the sky coordinates
    ! we need the shift of the left lower corner recorded in DETSEC
    if (grismFlag) then
       detsecValue=stringAttribute(inDataSet,"DETSEC")
       read(detsecValue,'(a1,i4,a6,i4)') auxc1,detsecx0,auxc6,detsecy0
    else
       detsecx0=0
       detsecy0=0
    endif
    exposureTime = real32Attribute(inDataSet, "EXPOSURE") ! Image exposure time

    !***********************************************************
    ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
    !***********************************************************
    call readModes(inDataSet, x0, y0, xsize, ysize)  
    oswImage => real32array2data(inArrayH)
 
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)
!
    trans(1) =  0.0_double
    trans(2) =  dcos(-rollAngle)
    trans(3) =  dsin(-rollAngle)
    trans(4) =  0.0_double
    trans(5) = -dsin(-rollAngle)
    trans(6) =  dcos(-rollAngle)

    !*******************************************
    ! create the output source file
    !*******************************************        
 
    PPSOSWDataSetH = DataSet(PPSOSWSet, create)

    !**********************************************************************
    ! Determine the bounding x and y (xmin, xmax, ymin, ymax) coordinates 
    ! of the image
    !*********************************************************************

    call  GetImageBoundaries(nx, ny, real(x0), real(y0), xBinning, yBinning, &
                             trans, xMin, xMax, yMin, yMax, grismFlag)

! invert transformation matrix for later use 
    call invert_trans(trans, itrans)    
! need inverted transformation matrix later to rotate image

    nxout = xmax - xmin                  ! determine image dimensions
    nyout = ymax - ymin

!********************************************************
!
! POSSIBLE BUG xmax & xmin are boundaries of the image
! in double precision
! nxout should be xmax - xmin + 1
!********************************************************

 
!***********************************************************************************
! set up the output image with the appropriate size (xmax,ymax already referenced to 
! origin of containing (output) box so these equal the output box dimensions).
!**********************************************************************************

    dim(1) = nxout     ! The dimensions of the output image
    dim(2) = nyout
    PPSOSWArrayH  = addArray(PPSOSWDataSetH, "PRIMARY",  Real32, dim)  ! For the image array


    ! copy the header across

    call copyAllAttributes(attributable(inDataSet), attributable(PPSOSWDataSetH))
   

    ! add content keyword

    call setAttribute(PPSOSWDataSetH, "CONTENT", "OM OSW FITS SKY IMAGE", &
         "Content of FITS file")

!********************************************
! convert from ra and dec numbers to strings
!********************************************

    call SLA_CR2TF(2, real(ra_point), cs, ihmsf)
    write(rastring, '(i2.2,'':'',i2.2,'':'',i2.2,''.'',i2.2)') &
         ihmsf(1), ihmsf(2), ihmsf(3), ihmsf(4)
    call SLA_CR2AF(1, real(dec_point), cs, idmsf)
    write(decstring, '(a1,i2.2,'':'',i2.2,'':'',i2.2,''.'',i1)') &
         cs, idmsf(1), idmsf(2), idmsf(3), idmsf(4)

!******************************************************************************
! now determine refDX, refDY offsets of OM plate centre (PC) from original OSW 
! origin (IN BINNED PIXELS)
!******************************************************************************
    if (grismFlag) then
       !
       refDX = omgXref  
       refDY = omgYref
       !
    else
       refDX = (xCentre - x0) / dble(Xbinning)
       refDY = (yCentre - y0) / dble(Ybinning)
    endif

! Compute these offsets in terms of the rotated frame (remember rotation 
! takes place about the OM plate centre so the x,y offsets of the OSW origin 
! shift in the rotated frame). 
    !*****************************
    ! NB NB NB temporary check

    ! determine the xmin, xmax, ymin, ymax of the spectra-aligned image
    xminr=99999.9
    xmaxr=-99999.9
    yminr=99999.9
    ymaxr=-99999.9
    !--------------------
    ! rotated lower left corner
    !-------------------------
    refDX=1.0_double
    refDY=1.0_double
    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 

    if (rotatedRefDX .lt. xminr) xminr=rotatedRefDX
    if (rotatedRefDX .gt. xmaxr) xmaxr=rotatedRefDX
    if (rotatedRefDY .lt. yminr) yminr=rotatedRefDY
    if (rotatedRefDY .gt. ymaxr) ymaxr=rotatedRefDY
    !---------------------------
    ! rotated lower right corner
    !----------------------------
    refDX=nx
    refDY=1.0_double
    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 

    if (rotatedRefDX .lt. xminr) xminr=rotatedRefDX
    if (rotatedRefDX .gt. xmaxr) xmaxr=rotatedRefDX
    if (rotatedRefDY .lt. yminr) yminr=rotatedRefDY
    if (rotatedRefDY .gt. ymaxr) ymaxr=rotatedRefDY
    !---------------------------
    ! rotated upper right corner
    !---------------------------
    refDX=nx
    refDY=ny
    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 

    if (rotatedRefDX .lt. xminr) xminr=rotatedRefDX
    if (rotatedRefDX .gt. xmaxr) xmaxr=rotatedRefDX
    if (rotatedRefDY .lt. yminr) yminr=rotatedRefDY
    if (rotatedRefDY .gt. ymaxr) ymaxr=rotatedRefDY
    !-----------------
    ! rotated upper left corner
    !------------------
    refDX=1.0_double
    refDY=ny
    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 

    if (rotatedRefDX .lt. xminr) xminr=rotatedRefDX
    if (rotatedRefDX .gt. xmaxr) xmaxr=rotatedRefDX
    if (rotatedRefDY .lt. yminr) yminr=rotatedRefDY
    if (rotatedRefDY .gt. ymaxr) ymaxr=rotatedRefDY
    !
    !-----------------------------------------------------
    ! compute the rotated position of the reference pixel
    !----------------------------------------------------
    refDX=omgXref
    refDY=omgYref
    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle

    rotatedRefDX=rotatedRefDX-xminr
    rotatedRefDY=rotatedRefDY-yminr

    ! NB NB NB 
    !****************************
    !---------------------------
    !compute two scales
    !------------------------

    scaleAlpha=abs(plateScaleCorrection*sinRollangle*degreesToRadians/3600.0)
    scaleDelta=abs(plateScaleCorrection*cosRollangle*degreesToRadians/3600.0)
    
    scaleAlpha=scaleAlpha+plateScale
    scaleDelta=scaleDelta+plateScale

! Finally compute offsets (refx, refy) of the OM plate centre from the BB origin, i.e. 
! just add the offset of the PC from the OSW origin to the offset of the OSW origin 
! from the BB origin
    rotateddsx0 = detsecx0 * cosRollangle - detsecy0 * sinRollangle
    rotateddsy0 = detsecy0 * sinRollangle + detsecy0 * cosRollangle 


    refx = rotatedRefDX
    refy = rotatedRefDY

!
! And write the WCS attributes to output file. Note that the scale is in binned pixels (ie 
! if binning is 2, with raw scale -0.5" per pixels, CDELT1/2 =1
!

    !****************************************************
    ! Set the fits header attributes for the image array
    !****************************************************
    call SetGrismFitsArrayAttributes(PPSOSWArrayH, xBinning, yBinning, scaleAlpha, &
         scaleDelta, refx, refy, ra_scx, dec_scx, ra_point, dec_point, grismFlag)

    outimage => real32array2data(PPSOSWArrayH)

    !*********************************************************  
    !  Now rotate the image.  
    !*********************************************************

    k = 0
    do i = 0, nxout - 1
        do j = 0, nyout - 1
            call setToNull( PPSOSWArrayH, k)

            k = k + 1
        end do
    end do

    ! Actually rotate the image (stored in outImage)
    ! VNY 23.02.2006 introduced the parameter grismFlag
    if (grismFlag) then
       call RotateGrismImage(nx, ny, nxout, nyout, oswImage, &
            outImage, x0, y0, trans, itrans, grismFlag)  
    else
       call RotateImage(nx, ny, nxout, nyout, oswImage, outImage, itrans, grismFlag)  
    endif
 
    ! Release handles

    call release(inArrayH)
    
    call release(PPSOSWArrayH) 
    
    call release(PPSOSWDataSetH)



    call release(inDataSet)

! Now check the header keywords CTYPE1 and CTYPE2 (only for the grim image)
    if (grismFlag) then
       PPSOSWDataSetH = DataSet(PPSOSWSet, modify)
       !PPSOSWArrayH  = array(PPSOSWDataSetH, "PRIMARY", modify) 
       if (hasAttribute(PPSOSWDataSetH, "CTYPE1")) then
          call deleteAttribute(PPSOSWDataSetH ,"CTYPE1")
          call setAttribute(PPSOSWDataSetH, "CTYPE1", "RA---TAN", "Coordinate system for X")
       endif!


       if (hasAttribute(PPSOSWDataSetH, "CTYPE2")) then
          call deleteAttribute(PPSOSWDataSetH,"CTYPE2")
          call setAttribute(PPSOSWDataSetH, "CTYPE2", "DEC--TAN", "Coordinate system for Y")
       endif
       !call release(PPSOSWArrayH) 
       call release(PPSOSWDataSetH)
    endif

end subroutine CreateGrismSkyImage


!***************************************************************
! Subroutine RotateImage
! Thius is done by-
! 1) looping over regular grid of output (sky) image points, 
! 2) computing (linear) location in unrotated image (via inverse transform)
! 3) correcting linear position in unrotated frame for distortion (i.e.  
!     converting back to raw pixel position) (2 stage iteration)
! 4) bi-linearly interpolating surrounding points 
!
!****************************************************************
! 23.02.2006 VNY introduced the parameter grismFlag
subroutine RotateImage(nx, ny, nxout, nyout, &
                      oswImage, rotatedImage, itrans, grismFlag)  
    integer, intent(in) :: nx, ny
  
    integer, intent(in) :: nxout, nyout
    real(single), pointer  :: oswImage(:, :)
    real(single), pointer :: rotatedImage(:, :)
    real(double), intent(in)  :: itrans(6)     ! its inverse
    logical(bool), intent(in) :: grismFlag
 
    !*******************************
    ! Local variables
    !*******************************
    real(double)  :: rx0, ry0, xDistorted, yDistorted
  
    real(double), dimension(:), pointer :: itrans3TimesjPlusPoint5  ! Stores itrans(3) * (j + 0.5)
    real(double), dimension(:), pointer :: itrans6TimesjPlusPoint5  ! Stores itrans(6) * (j + 0.5)
    real(single) :: pixelValue    
    integer :: i, j, istat 
    real(double), parameter :: shift = 1.0   
    real(single) :: minImage, maxImage
    

    minImage = minval(oswImage)
    maxImage = maxval(oswImage)

    write(errstr, '(a)') "Rotating the PRIMARY image"
    call message(AppMsg, Verbose)

    itrans3TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans3TimesjPlusPoint5")
    itrans6TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans6TimesjPlusPoint5")


    !***************************************************************
    ! The following is to reduce execution time when creating the
    ! rotated image
    !***************************************************************
    forall (j = 0 : nyout - 1) 

        itrans3TimesjPlusPoint5(j) = itrans(3) * (dble(j) + shift)
        itrans6TimesjPlusPoint5(j) = itrans(6) * (dble(j) + shift)

    end forall


    dblNx = dble(nx)
    dblNy = dble(ny)
    do i = 0, nxout - 1

       rx0 = itrans(1) + itrans(2) * (dble(i) + shift)
       ry0 = itrans(4) + itrans(5) * (dble(i) + shift)

     
       do j = 0, nyout - 1
        
          xOrigin = rx0 + itrans3TimesjPlusPoint5(j)   ! X position in undistorted image (centre of pixel)
          yOrigin = ry0 + itrans6TimesjPlusPoint5(j)   ! Y position in undistorted image (centre of pixel)


          !*****************************************************
          ! Get distorted position
          !*****************************************************
          ! VNY 23.02.2006 No distortion is added in the case of the grism filter
          if (grismFlag) then
             xDistorted=xOrigin
             yDistorted=yOrigin
          else
             call AddDistortion(xOrigin, yOrigin, xDistorted, yDistorted)
             xDistorted = xDistorted - 0.5
             yDistorted = yDistorted - 0.5
          endif
 
          if (xDistorted .gt. 0.0_double .and. xDistorted .lt. dblNx .and. & 
              yDistorted .gt. 0.0_double .and. yDistorted .lt. dblNy) then


              pixelValue = Interpolate(oswImage, &
                    real(xDistorted), real(yDistorted))

              if (pixelValue .ge. minImage .and. pixelValue .le. maxImage)  &
                   rotatedImage(i, j) = pixelValue

          end if                 
       end do
    end do

    


    deallocate(itrans3TimesjPlusPoint5, itrans6TimesjPlusPoint5, stat = istat)
    call  CheckDEAllocation(istat, "itrans3/6TimesjPlusPoint5")
 
end subroutine RotateImage


!***************************************************************
! Subroutine RotateGrismImage
! Thius is done by-
! 1) looping over regular grid of output (sky) image points, 
! 2) computing (linear) location in unrotated image (via inverse transform)
! 3) correcting linear position in unrotated frame for distortion (i.e.  
!     converting back to raw pixel position) (2 stage iteration)
! 4) bi-linearly interpolating surrounding points 
!
!****************************************************************
! 18.10.2006 VNY 
subroutine RotateGrismImage(nx, ny, nxout, nyout, oswImage, &
     rotatedImage, x0, y0, trans, itrans, grismFlag)  
    integer, intent(in) :: nx, ny
  
    integer, intent(in) :: nxout, nyout
    real(single), pointer  :: oswImage(:, :)
    real(single), pointer :: rotatedImage(:, :)
    integer, intent(in) :: x0, y0
    real(double), intent(in)  :: trans(6)
    real(double), intent(in)  :: itrans(6)     ! its inverse
    logical(bool), intent(in) :: grismFlag
 
    !*******************************
    ! Local variables
    !*******************************
    real(double)  :: rx0, ry0, xDistorted, yDistorted
  
    real(double), dimension(:), pointer :: itrans3TimesjPlusPoint5  ! Stores itrans(3) * (j + 0.5)
    real(double), dimension(:), pointer :: itrans6TimesjPlusPoint5  ! Stores itrans(6) * (j + 0.5)
    real(single) :: pixelValue    
    integer :: i, j, istat 
    real(double), parameter :: shift = 1.0   
    real(single) :: minImage, maxImage
    real(double) :: rotatedRefDX, rotatedRefDY, testX, testY

    minImage = minval(oswImage)
    maxImage = maxval(oswImage)

    write(errstr, '(a)') "Rotating the grism image"
    call message(AppMsg, Verbose)
 
    itrans3TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans3TimesjPlusPoint5")
    itrans6TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans6TimesjPlusPoint5")


    !***************************************************************
    ! The following is to reduce execution time when creating the
    ! rotated image
    !***************************************************************
    forall (j = 0 : nyout - 1) 

        !itrans3TimesjPlusPoint5(j) = itrans(3) * (dble(j) + shift)
        !itrans6TimesjPlusPoint5(j) = itrans(6) * (dble(j) + shift)
        itrans3TimesjPlusPoint5(j) = itrans(3) * (dble(j) + shift)
        itrans6TimesjPlusPoint5(j) = itrans(6) * (dble(j) + shift)

    end forall

    !----------------------------
    !NB NB NB 
    ! temporal test back-rotation of the image 18.10.2006
    i=0
    j=-710+2305
    rx0=itrans(2)*(dble(i))+itrans(3)*(dble(j))
    ry0=itrans(5)*(dble(i))+itrans(6)*(dble(j))
    rotatedRefDX=-1229.5_double
    rotatedRefDY=-710.2_double
    rotatedRefDX=rotatedRefDX+trans(1)
    rotatedRefDY=rotatedRefDY+trans(4)
    call Rotate(itrans,rotatedRefDX, rotatedRefDY, testX, testY)

    ! NB NB NB 
    !--------------------
    dblNx = dble(nx)
    dblNy = dble(ny)
    do i = 0, nxout - 1
       rx0 = itrans(1)+itrans(2) * (dble(i) + shift)
       ry0 = itrans(4)+itrans(5) * (dble(i) + shift)

       do j = 0, nyout - 1
        
          xOrigin = rx0 + itrans3TimesjPlusPoint5(j)   
! X position in undistorted image (centre of pixel)
          yOrigin = ry0 + itrans6TimesjPlusPoint5(j)   
! Y position in undistorted image (centre of pixel)
!          if (i .eq. 434 .and. j .eq. 1737) then
!             write(*,*) "L900 i=", i, " j=", j, " xOrigin=", xOrigin, &
!                  " yOrigin=", yOrigin
!             rotatedRefDX=dble(i)
!             rotatedRefDY=dble(j)
!             call Rotate(itrans, rotatedRefDX, rotatedRefDY, testX, testY)
!             write(*,*) "L905 rotatedRefDX=", rotatedRefDX, " rotatedRefDY=", rotatedRefDY, &
!                  " testX=", testX, " testY=", testY   
!          endif

          !*****************************************************
          ! Get distorted position
          !*****************************************************
          ! VNY 23.02.2006 No distortion is added in the case of the grism filter
          if (grismFlag) then
             xDistorted=xOrigin-0.5 -dble(x0)
             yDistorted=yOrigin-0.5 -dble(y0)
          else
             call AddDistortion(xOrigin, yOrigin, xDistorted, yDistorted)
             xDistorted = xDistorted - 0.5
             yDistorted = yDistorted - 0.5
          endif
 
          if (xDistorted .gt. 0.0_double .and. xDistorted .lt. dblNx .and. & 
              yDistorted .gt. 0.0_double .and. yDistorted .lt. dblNy) then


              pixelValue = Interpolate(oswImage, &
                    real(xDistorted), real(yDistorted))

              if (pixelValue .ge. minImage .and. pixelValue .le. maxImage)  &
                   rotatedImage(i, j) = pixelValue

          end if                 
       end do
    end do

    
    deallocate(itrans3TimesjPlusPoint5, itrans6TimesjPlusPoint5, stat = istat)
    call  CheckDEAllocation(istat, "itrans3/6TimesjPlusPoint5")
 
end subroutine RotateGrismImage


!*******************************************************************************
! subroutine GetImageBoundaries
! The first step is to find the limits of the sky box that will bound the rotated 
! image, after it is linearised (distortion corrections applied). This is done by 
!
! 1) computing the raw coordinates of the image corners and applying the appropriate 
! distortion correction to them. The coordinates are referenced back to the origin 
! of the undistorted image. The corners are then rotated by the roll angle and 
! converted to binned pixels 
! 2) the limits of this rotated, distorted image are obtained and the translational 
! shift betwen the bounding box origin and the OSW origin are passed to the 
! transformation matrix leading coefficients.
! 3) The dimensions of the bounding box are determined
! 4) The offset of the OM plate centre (1024,1024) are determined relative to 
! the OSW original, as measured in the rotated frame, and then this is 
! corrected to be wrt to the bounding box origin
! 5) WCS keywords are set
! 6) pixels in the sky box are now converted, via the inverse transformation matrix, 
! to the OSW frame and an iterative process to correct back from linearised to raw 
! pixels is used to find the original pixel in the raw image. Bi-linear interpolation 
! is used to fill the rotated image  
!*******************************************************************************
subroutine GetImageBoundaries(nx, ny, x0, y0, xBinning, yBinning, &
                          trans, xMin, xMax, yMin, yMax, grismFlag)
    integer, intent(in) :: nx, ny   ! The image dimensions
    real(single),  intent(in) :: x0, y0
    integer, intent(in) :: xBinning, yBinning
    real(double), intent(inout), dimension(1 : 6) :: trans   
    real(double), intent(out) :: xMin, xMax, yMin, yMax
    real(double)  :: xlin(4), ylin(4), xcorn(4), ycorn(4)
    real(double) :: dy
    logical(bool), intent(in) :: grismFlag
     
    ! bottom left hand corner - at 2048 pixels resolution 

    ! correct for distortion and reference to x0,y0 origin
    if (grismFlag) then
       xLin(1)=x0
       yLin(1)=y0
    else
       call RemoveDistortion1(x0, y0, xlin(1), ylin(1))
    endif


    ! rotate + convert to binned pixels
    call Rotate(trans, xlin(1), ylin(1), xCorn(1), yCorn(1))
    ! bottom rh corner in 2048 pixels
    if (grismFlag) then
       xLin(2)=x0+real(nx*xBinning)
       yLin(2)=y0
    else
       call RemoveDistortion1(x0 + real(nx * xBinning), y0, xlin(2), ylin(2))
    endif

    call Rotate(trans, xlin(2), ylin(2), xCorn(2), yCorn(2))

    ! top rh corner
    if (grismFlag) then
       xLin(3)=x0+real(nx*xBinning)
       yLin(3)=y0+real(ny*yBinning)
    else
       call RemoveDistortion1(x0 + real(nx * xBinning), y0 + real(ny * yBinning), xlin(3), ylin(3))
    endif
    call Rotate(trans, xlin(3), ylin(3), xCorn(3), yCorn(3))

    ! top lh corner
    if (grismFlag) then
       xLin(4)=x0
       yLin(4)=y0+real(ny*yBinning)
    else
       call RemoveDistortion1(x0, y0 + real(ny * yBinning), xlin(4), ylin(4))  
    endif

    call Rotate(trans, xlin(4), ylin(4), xCorn(4), yCorn(4))


    !  We now have the corners of OSW in linearised coordinates relative to the origin 
    !  of the original OSW image, after rotation of the frame 
    !  Now need to get limits of bounding box (BB)  from these corners (in rotated frame)
    !  Once minimum x and y are found (origin of BB - i.e. the rotated frame), since these 
    !  are wrt the original OSW origin, we can determine the offset of the OSW origin from 
    !  bounding BB (transfer offsets to trans(1), trans(4))

    if (grismFlag) then
       xmin=999999.0
       xmax=-999999.0
    else
       xmax = trans(1)
       xmin = trans(1)
    endif

    call GetMinMax(xmin, xmax, xcorn)

    ! Ensure xmax - xmin is an integral number of pixels
    ! and xmin is x.5
    
    trans(1) = trans(1) - xmin     ! get x offset of OSW origin from BB origin

    if (grismFlag) then
       ymin=999999.0
       ymax=-999999.0
    else
       ymax = trans(4)
       ymin = trans(4)
    endif
    call GetMinMax(ymin, ymax, ycorn)

    ! Ensure ymax - ymin is an integral number of pixels
    dy = ymax - ymin
    dy = real(ceiling(ymax - ymin))
    ymin = real(floor(ymin)) + 0.5
    ymax = ymin + dy

    trans(4) = trans(4) - ymin     ! get y offset of OSW origin from BB origin
end subroutine GetImageBoundaries
!******************************
! subrioutine GetMinMax
!******************************
subroutine GetMinMax(minimum, maximum, corner)
    real(double), intent(inout) :: minimum
    real(double), intent(inout) :: maximum
    real(double), intent(in), dimension(1 : 4) :: corner
    integer :: i

    do i = 1, 4
        minimum = dmin1(minimum, corner(i))
        maximum = dmax1(maximum, corner(i))
    end do

end subroutine GetMinMax 
!*****************************************
! subroutine SetFitsArrayAttributes
!*****************************************
subroutine SetFitsArrayAttributes(arrayH, xBinning, yBinning,     &
                                  plateScale, refx, refy, ra_scx, &
                                  dec_scx, ra_point, dec_point, &
                                  grismFlag, posCorr, ra_off, dec_off)

    type(ArrayT), intent(in) :: arrayH   ! Array handle.
    integer, intent(in) :: xBinning, yBinning 
    real(double), intent(in) :: plateScale
    real(single), intent(in) :: refx, refy
    real(double), intent(in) :: ra_scx, dec_scx
    real(double), intent(in) :: ra_point, dec_point
    logical(bool), intent(in) :: grismFlag
    logical(bool), intent(in) :: posCorr
    real(single), intent(in) :: ra_off, dec_off


    if (.not. grismFlag) then
       call setAttribute(arrayH, "CTYPE1", "RA---TAN", "Coordinate system for X")
    endif

    call setAttribute(arrayH, "CRPIX1", refx, "pix", "WCS reference pixel") 
    call setAttribute(arrayH, "RA_SCX", ra_scx * radiansToDegrees, "deg", &
                              "Coordinate of spacecraft pointing direction for X")
    call setAttribute(arrayH, "CRVAL1", ra_point * radiansToDegrees, "deg", &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "CDELT1", -real(plateScale * dble(xBinning) * radiansToDegrees), &
                              "deg", "WCS pixel size")
    if (.not. grismFlag) then
       call setAttribute(arrayH, "CTYPE2", "DEC--TAN", "Coordinate system for Y")
    endif

    call setAttribute(arrayH, "CRPIX2", refy, "pix", "WCS reference pixel")

    call setAttribute(arrayH, "DEC_SCX", dec_scx * radiansToDegrees, &
                              "deg", "Coordinate of spacecraft pointing direction for Y")
    call setAttribute(arrayH, "CRVAL2", dec_point * radiansToDegrees, &
                              "deg", "Coordinate of pointing direction for Y")
    call setAttribute(arrayH, "CDELT2", real(plateScale * dble(yBinning) * radiansToDegrees), &
                              "deg", "WCS pixel size")
    call setAttribute(arrayH, "RA_PNT", ra_point * radiansToDegrees, "deg", &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "DEC_PNT", dec_point * radiansToDegrees, "deg", &
                              "Coordinate of pointing direction for Y")
    if (posCorr) then
       call setAttribute(arrayH, "POSCOROK", .true., &
            "Astrometry correction OK")
       call setAttribute(arrayH, "RA_OFF", ra_off, "arcsec", &
            "Astrom. correction applied to CRVAL1")
       call setAttribute(arrayH, "DEC_OFF", dec_off, "arcsec", &
            "Astrom. correction applied to CRVAL2")

       !************************************************************************************************
       !
       ! January 2009- to avoid confusion add the same astrometric keywords that are found in
       !               the SWS file
       !
       !***********************************************************************************************



       call setAttribute(arrayH, "RAOFFSET", ra_off * cos(dec_point), "arcsec", "X offset at pointing declination")
       call setAttribute(arrayH, "DEOFFSET", dec_off, "arcsec", "Y offset at pointing declination")
       call addcomment(arrayH, "NB to obtain the actual RA/DEC corrections applied divide the offsets by cos(dec)")




    else
       call setAttribute(arrayH, "POSCOROK", .false., &
            "Astrometry correction OK")
       call setAttribute(arrayH, "RA_OFF", 0.0, "arcsec", &
            "Astrom. correction applied to CRVAL1")
       call setAttribute(arrayH, "DEC_OFF", 0.0, "arcsec",  &
            "Astrom. correction applied to CRVAL2")

       call setAttribute(arrayH, "RAOFFSET", 0.0, "arcsec", "X offset at pointing declination")
       call setAttribute(arrayH, "DEOFFSET", 0.0, "arcsec", "Y offset at pointing declination")
       call addcomment(arrayH, "NB to obtain the actual RA/DEC corrections applied divide the offsets by cos(dec)")


    endif
 
end subroutine SetFitsArrayAttributes

!*****************************************
! subroutine SetFitsArrayAttributes
!*****************************************
subroutine SetGrismFitsArrayAttributes(arrayH, xBinning, yBinning,     &
                                  plateAlpha, plateDelta, refx, refy, ra_scx, &
                                  dec_scx, ra_point, dec_point,grismFlag)

    type(ArrayT), intent(in) :: arrayH   ! Array handle.
    integer, intent(in) :: xBinning, yBinning 
    real(double), intent(in) :: plateAlpha, plateDelta
    real(single), intent(in) :: refx, refy
    real(double), intent(in) :: ra_scx, dec_scx
    real(double), intent(in) :: ra_point, dec_point
    logical(bool), intent(in) :: grismFlag


    if (.not. grismFlag) then
       call setAttribute(arrayH, "CTYPE1", "RA---TAN", "Coordinate system for X")
    endif

    call setAttribute(arrayH, "CRPIX1", refx, "WCS reference pixel") 
    call setAttribute(arrayH, "RA_SCX", ra_scx * radiansToDegrees, &
                              "Coordinate of spacecraft pointing direction for X")
    call setAttribute(arrayH, "CRVAL1", ra_point * radiansToDegrees, &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "CDELT1", -real(plateAlpha * dble(xBinning) * radiansToDegrees), &
                              "WCS pixel size")
    if (.not. grismFlag) then
       call setAttribute(arrayH, "CTYPE2", "DEC--TAN", "Coordinate system for Y")
    endif

    call setAttribute(arrayH, "CRPIX2", refy, "WCS reference pixel")

    call setAttribute(arrayH, "DEC_SCX", dec_scx * radiansToDegrees, &
                              "Coordinate of spacecraft pointing direction for Y")
    call setAttribute(arrayH, "CRVAL2", dec_point * radiansToDegrees, &
                              "Coordinate of pointing direction for Y")
    call setAttribute(arrayH, "CDELT2", real(plateDelta * dble(yBinning) * radiansToDegrees), &
                              "WCS pixel size")
    call setAttribute(arrayH, "RA_PNT", ra_point * radiansToDegrees, &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "DEC_PNT", dec_point * radiansToDegrees, &
                              "Coordinate of pointing direction for Y")
 
end subroutine SetGrismFitsArrayAttributes


!*****************************************
! subroutine SetFitsTableAttributes
!*****************************************
subroutine SetFitsTableAttributes(arrayH, ra_point, dec_point)

    type(DataSetT), intent(in) :: arrayH   ! Array handle.
     
    
    real(double), intent(in) :: ra_point, dec_point



    call setAttribute(arrayH, "RA_PNT", ra_point * radiansToDegrees, &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "DEC_PNT", dec_point * radiansToDegrees, &
                              "Coordinate of pointing direction for Y")
 
end subroutine SetFitsTableAttributes


!******************************************
! invert_trans
! inverts the transformation matrix, i.e. gets coefficients to swap between
! x'= t(1) + t(2)*x + t(3)*y
! y'= t(4) + t(5)*x + t(6)*y
! and 
! x= it(1) + it(2)*x' + it(3)*y'
! y= it(4) + it(5)*x' + it(6)*y'
!******************************************
subroutine invert_trans(trans, itrans)
    real(double), intent(in) ::  trans(6)
    real(double), intent(out) :: itrans(6)
    real(double) determ

    determ = trans(2) * trans(6) - trans(3) * trans(5)
    itrans(1) = trans(3) * trans(4) - trans(1) * trans(6)
    itrans(2) = trans(6)
    itrans(3) = -trans(3)
    itrans(4) = trans(1) * trans(5) - trans(2) * trans(4)
    itrans(5) = -trans(5)
    itrans(6) = trans(2)
    itrans = itrans / determ

end subroutine invert_trans

subroutine invert_trans_grism(trans, itrans)
    real(double), intent(in) ::  trans(6)
    real(double), intent(out) :: itrans(6)
    real(double) determ

    determ = trans(2) * trans(6) - trans(3) * trans(5)

    itrans(2) = trans(6)
    itrans(3) = -trans(3)

    itrans(5) = -trans(5)
    itrans(6) = trans(2)
    itrans = itrans / determ

    itrans(1) = -trans(1) * itrans(2)  - trans(4) * itrans(3)
    itrans(4) = -trans(1) * itrans(5)  - trans(4) * itrans(6)

end subroutine invert_trans_grism

!************************************************************
! 
!***********************************************************
subroutine Rotate(trans, xIn, yIn, xOut, yOut)
    real(double), intent(in) :: trans(:)
    real(double), intent(in) :: xIn, yIn
    real(double), intent(out) :: xOut, yOut

    xOut = (trans(1) + trans(2) * xIn + trans(3) * yIn) / dxBinning 
    yOut = (trans(4) + trans(5) * xIn + trans(6) * yIn) / dyBinning
end subroutine Rotate



!***********************************************************
!
!***********************************************************
real(single) Function Interpolate(image, x, y)
    real(single), pointer :: image(:, :)
    real(single), intent(in) :: x, y

    real(single) :: xl, xu, yl, yu
    real(single) :: u, t
    integer j0, k0, j1, k1
    integer nx, ny

    nx = size(image, 1)
    ny = size(image, 2)
    xl = 0.0
    yl = 0.0
    xu = real(nx)
    yu = real(ny)

    if (x .gt. xl .and. x .lt. xu .and. & 
        y .gt. yl .and. y .lt. yu) then

        

        j0 = max(0, int(x - 0.5))
        j0 = min(nx - 1, j0)
      
        k0 = max(0, int(y - 0.5))
        k0 = min(ny - 1, k0)


        
        j1 = j0 + 1
        k1 = k0 + 1

        if (j1 .ge. nx) then
            t = 0.0
            j1 = nx - 1
        else        
            t = max(0.0, x - real(j0) - 0.5)
        end if
        if (k1 .ge. ny) then
            k1 = ny - 1
            u = 0.0
        else  
            u = max(0.0, y - real(k0) - 0.5)
        end if
      
        Interpolate = &
              (1.0 - t) * (1.0 - u) * image(j0, k0) + &
               t * (1.0 - u) * image(j1, k0) + &
               t * u * image(j1, k1) + &
               (1.0  - t) * u * image(j0, k1)



        

        else
            Interpolate = -1.0
    end if


end function Interpolate





!*****************************************************************
! July 20 2008 Adding the quality image to the sky image
!
! AddQualityImage
!
!
! 
!****************************************************************
subroutine addQualityImage(set, PPSOSWSET, rollAngle, ra_point, dec_point, &
              xBinning, yBinning, ra_scx, dec_scx, plateScale, grismFlag)
    character(*), intent(in) :: set                                         ! OM image intermediate filename
 
    character(*), intent(in) :: PPSOSWSet    ! PPS OM OSW FITS sky image filename.

    real(double), intent(in) :: rollAngle
    real(double), intent(in) :: ra_point, dec_point        ! position of the OM tangent point
    
    integer, intent(in)        :: xBinning
    integer, intent(in)        :: yBinning
    real(double), intent(in) :: ra_scx, dec_scx  ! The ra, dec and pa of the spacecraft     
    real(double) :: ra_scx2, dec_scx2
    real(double), intent(in) :: plateScale         ! SRR added 4/10/00
    real(double) :: plateScaleAux         ! SRR added 4/10/00
    logical(bool), intent(in) :: grismFlag

    real(single) :: refx, refy
    character(1) :: auxc1
    character(6) :: auxc6


    type(DataSetT) :: inDataSet                                       ! Input dataset handle.


 
    type(DataSetT) :: PPSOSWDataSetH                                  ! PPS OSW dataset handle.

    type(ArrayT) :: PPSOSWArrayH                                      ! PPS OSW array handle.


    real(double)  :: trans(6) ! rotation transformation matrix
    real(double)  :: itrans(6)     ! its inverse

    integer(int16), pointer :: oswImage(:, :) => null()
 
    integer(int16), pointer  :: outimage(:,:) => null()      ! The output image
    
  
    real(double)  :: sinRollAngle, cosRollAngle
    integer             :: idmsf(4), ihmsf(4)  
    integer             :: nxout, nyout        ! size of the output image
    integer                    :: dim(2), nx, ny     ! size of the image

    real(single)                       :: refDX
    real(single)                       :: refDY
    real(single)                       :: rotatedRefDX
    real(single)                       :: rotatedRefDY
    real(double)  :: xmax,xmin,ymax,ymin            
    integer(int32) ::  i, j, k    
    character(20):: rastring, decstring   ! strings for RA and DEC
    character(1) :: cs                    ! sign of dec +/-
    integer             :: x0, y0             ! offset of the image
    integer             :: xsize, ysize       ! size of the image

    !character(6)            :: omFilter   ! the OM filter
    character(21) :: detsecValue
   
    integer :: detsecx0, detsecy0, rotateddsx0, rotateddsy0
   

    plateScaleAux=PlateScale
    ra_scx2=ra_scx
    dec_scx2=dec_scx
  
    write(errstr, '(a)') "Adding the QUALITY image"
    call message(AppMsg, Verbose)


 

    xBin = xBinning
    yBin = yBinning
    sinRollAngle = dsin(rollAngle)  ! Need these quantities quite a lot
    cosRollAngle = dcos(rollAngle)
  
 
    inDataSet = DataSet(set, read) 


    call GetQualityImage( inDataSet, xSize, ySize, oswImage) 


 

    !***********************************************************
    ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
    !***********************************************************
    call readModes(inDataSet, x0, y0, xsize, ysize)  
   
 
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)
    trans(1) =  0.0_double
    trans(2) =  dcos(-rollangle)
    trans(3) =  dsin(-rollangle)
    trans(4) =  0.0_double
    trans(5) = -dsin(-rollangle)
    trans(6) =  dcos(-rollangle)

    !*********************************************************
    ! modify the output source file- add the QUALITY array
    !*********************************************************        
 
    PPSOSWDataSetH = DataSet(PPSOSWSet, modify)

    !**********************************************************************
    !
    ! Determine the bounding x and y (xmin, xmax, ymin, ymax) coordinates 
    ! of the image
    !
    !*********************************************************************
    call  GetImageBoundaries(nx, ny, real(x0), real(y0), xBinning, yBinning, &
                             trans, xMin, xMax, yMin, yMax, grismFlag)

! invert transformation matrix for later use
    call invert_trans(trans, itrans)    
! need inverted transformation matrix later to rotate image

    nxout = xmax - xmin                  ! determine image dimensions
    nyout = ymax - ymin



!********************************************************
!
! POSSIBLE BUG xmax & xmin are boundaries of the image
! in double precision
! nxout should be xmax - xmin + 1
!********************************************************

 
    !***********************************************************************************
    ! set up the output image with the appropriate size (xmax,ymax already referenced to 
    ! origin of containing (output) box so these equal the output box dimensions).
    !**********************************************************************************

    dim(1) = nxout     ! The dimensions of the output image
    dim(2) = nyout
    PPSOSWArrayH  = addArray(PPSOSWDataSetH, "QUALITY",  Integer16, dim)  ! For the QUALITY array



    ! in order to transform correctly the grism image into the sky coordinates
    ! we need the shift of the left lower corner recorded in DETSEC
    if (grismFlag) then
       detsecValue=stringAttribute(inDataSet,"DETSEC")
       read(detsecValue,'(a1,i4,a6,i4)') auxc1,detsecx0,auxc6,detsecy0
    else
       detsecx0=0
       detsecy0=0
    endif



    !********************************************
    !
    ! convert from ra and dec numbers to strings
    !
    !********************************************

    call SLA_CR2TF(2, real(ra_point), cs, ihmsf)
    write(rastring, '(i2.2,'':'',i2.2,'':'',i2.2,''.'',i2.2)') &
         ihmsf(1), ihmsf(2), ihmsf(3), ihmsf(4)
    call SLA_CR2AF(1, real(dec_point), cs, idmsf)
    write(decstring, '(a1,i2.2,'':'',i2.2,'':'',i2.2,''.'',i1)') &
         cs, idmsf(1), idmsf(2), idmsf(3), idmsf(4)

    !******************************************************************************
    ! now determine refDX, refDY offsets of OM plate centre (PC) from original OSW 
    ! origin (IN BINNED PIXELS)
    !******************************************************************************
    refDX = (xCentre - x0) / dble(Xbinning)
    refDY = (yCentre - y0) / dble(Ybinning)
   

    !****************************************************************************
    ! Compute these offsets in terms of the rotated frame (remember rotation 
    ! takes place about the OM plate centre so the x,y offsets of the OSW origin 
    ! shift in the rotated frame). 
    !*****************************************************************************
    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 


    !**************************************************************************************
    ! Finally compute offsets (refx, refy) of the OM plate centre from the BB origin, i.e. 
    ! just add the offset of the PC from the OSW origin to the offset of the OSW origin 
    ! from the BB origin
    !************************************************************************************ 
    rotateddsx0 = detsecx0 * cosRollangle - detsecy0 * sinRollangle
    rotateddsy0 = detsecy0 * sinRollangle + detsecy0 * cosRollangle 

    refx = trans(1) + rotatedRefDX    
    refy = trans(4) + rotatedRefDY
    

   
!
! And write the WCS attributes to output file. Note that the scale is in binned pixels (ie 
! if binning is 2, with raw scale -0.5" per pixels, CDELT1/2 =1
!

    !****************************************************
    ! Set the fits header attributes for the image array
    !****************************************************
!    call SetFitsArrayAttributes(PPSOSWArrayH, xBinning, yBinning, plateScale, refx, refy, &
!                                  ra_scx, dec_scx, ra_point, dec_point, grismFlag, posCorr, &
!                                  ra_off, dec_off)

    outimage => int16array2data(PPSOSWArrayH)

    !*********************************************************  
    !  Now rotate the image.  
    !*********************************************************

   


    call setNullValue( PPSOSWarrayH, -999)
    k = 0
    do i = 0, nxout - 1
        do j = 0, nyout - 1
            call setToNull( PPSOSWArrayH, k)

            k = k + 1
        end do
    end do


    !***********************************************************
    ! Actually rotate the image (stored in outImage)
    ! VNY 23.02.2006 introduced the parameter grismFlag
    !***********************************************************
    call RotateQualityImage(nx, ny, nxout, nyout, oswImage, outImage, itrans, grismFlag)  

 
    ! Release handles


    
    call release(PPSOSWDataSetH)
    call release(inDataSet)
    deallocate( oswImage )
    nullify( oswImage )

end subroutine addQualityImage

!*******************************************
! subroutine GetImage
! Gets the 2-d image from the fits file
!*******************************************
subroutine GetQualityImage(dataSet, nx, ny, image) 

    type(dataSetT) :: dataSet
 
    integer(int16), pointer, dimension(: , :) :: image
    integer(int32), intent(out) :: nx, ny    
    integer(int8), dimension(:, :),  pointer :: int8ImageP     ! Integer8 array pointer.
    integer(int16), dimension(:, :), pointer :: int16ImageP    ! Integer16 array pointer.
    
    integer :: dim(2)
    type(ArrayT) :: inArray
    integer :: nDim
    
    inArray = array(dataSet, "QUALITY", read)
!
    nDim  = numberOfDimensions(inArray)             ! How many dimensions?
    ! If it's not an image then crash out
    if (nDim .ne. 2) call fatal('notImage','input file not an image')
    ! Get the number of bins.
    dim = dimensions(inArray)
    nx = dim(1)
    ny = dim(2)
    if (nx <= 0) call fatal("Bad image", "nx <=0")
    if (ny <= 0) call fatal("Bad image", "ny <=0")




    if (associated(image)) deallocate(image)
    image => Allocate2DInt16(0,  nx - 1,  0 , ny - 1, "image")
    ! Convert the image to real, if necessary
    select case(arrayDataType(inArray))
        case(integer8)   
            int8ImageP => int8Array2Data(inArray)
            image = int(int8ImageP, int16)
        case(integer16)
            int16ImageP => int16Array2Data(inArray)
            image = int16ImageP


    end select
    call release(inArray)


end subroutine GetQualityImage







!***************************************************************
! Subroutine RotateQualityImage
! Thius is done by-
! 1) looping over regular grid of output (sky) image points, 
! 2) computing (linear) location in unrotated image (via inverse transform)
! 3) correcting linear position in unrotated frame for distortion (i.e.  
!     converting back to raw pixel position) (2 stage iteration)
! 4) bi-linearly interpolating surrounding points 
!
!****************************************************************

subroutine RotateQualityImage(nx, ny, nxout, nyout, &
                      oswImage, rotatedImage, itrans, grismFlag)  
    integer, intent(in) :: nx, ny
  
    integer, intent(in) :: nxout, nyout
    integer(int16), pointer  :: oswImage(:, :)
    integer(int16), pointer :: rotatedImage(:, :)
    real(double), intent(in)  :: itrans(6)     ! its inverse
    logical(bool), intent(in) :: grismFlag
 
    !*******************************
    ! Local variables
    !*******************************
    real(double)  :: rx0, ry0, xDistorted, yDistorted
  
    real(double), dimension(:), pointer :: itrans3TimesjPlusPoint5  ! Stores itrans(3) * (j + 0.5)
    real(double), dimension(:), pointer :: itrans6TimesjPlusPoint5  ! Stores itrans(6) * (j + 0.5)
   
    integer :: i, j, k, l, istat, m, n, o 
    real(double), parameter :: shift = 1.0   
    integer(int16) :: minImage, maxImage, pixelValue
  

    minImage = minval(oswImage)
    maxImage = maxval(oswImage)

    write(errstr, '(a)') "Rotating the QUALITY image"
    call message(AppMsg, Verbose)

    itrans3TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans3TimesjPlusPoint5")
    itrans6TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans6TimesjPlusPoint5")


    !***************************************************************
    ! The following is to reduce execution time when creating the
    ! rotated image
    !***************************************************************
    forall (j = 0 : nyout - 1) 

        itrans3TimesjPlusPoint5(j) = itrans(3) * (dble(j) + shift)
        itrans6TimesjPlusPoint5(j) = itrans(6) * (dble(j) + shift)

    end forall


    dblNx = dble(nx)
    dblNy = dble(ny)

    do i = 0, nxout - 1

       rx0 = itrans(1) + itrans(2) * (dble(i) + shift)
       ry0 = itrans(4) + itrans(5) * (dble(i) + shift)

     
       do j = 0, nyout - 1
        
          xOrigin = rx0 + itrans3TimesjPlusPoint5(j)   ! X position in undistorted image (centre of pixel)
          yOrigin = ry0 + itrans6TimesjPlusPoint5(j)   ! Y position in undistorted image (centre of pixel)


          !*****************************************************
          ! Get distorted position
          !*****************************************************
         
          if (grismFlag) then
             xDistorted = xOrigin
             yDistorted = yOrigin
          else
             call AddDistortion(xOrigin, yOrigin, xDistorted, yDistorted)
             xDistorted = xDistorted - 0.5
             yDistorted = yDistorted - 0.5
          endif
 
          if (xDistorted .gt. 0.0_double .and. xDistorted .lt. dblNx .and. & 
              yDistorted .gt. 0.0_double .and. yDistorted .lt. dblNy) then

              k = nint(xDistorted)
              k = max(0, k)
              k =  min(k, nx - 1)
              l = nint(yDistorted)
              l = max(0, l)
              l =  min(l, ny - 1)

              pixelValue = oswImage(k, l)              

              do m = max(0, k - 1), min(nx - 1, k + 1)
                  do n = max(0, l - 1), min(ny - 1, l + 1)
                      do o = 0, 10
                          if ( btest(oswImage(m, n), o) ) then
                              pixelValue = ibset( pixelValue, o) 
                          end if
                      end do
                  end do
              end do

              rotatedImage(i, j) = pixelValue

          end if                 
       end do

    end do
    
    deallocate(itrans3TimesjPlusPoint5, itrans6TimesjPlusPoint5, stat = istat)
    call  CheckDEAllocation(istat, "itrans3/6TimesjPlusPoint5")
 
end subroutine RotateQualityImage

end module omrotation_mod













































































